# Copyright Toyota Research Institute 2019
"""
Module containing basic agent Abstractions
and functionality
"""

import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import KFold, cross_val_score
from sklearn.base import clone

from camd import tqdm

import abc


class HypothesisAgent(metaclass=abc.ABCMeta):
    """
    Abstract class for agents, decision-making entities
    in a sequential learning setting.  Should implement
    a `get_hypotheses` method that takes a list of potential
    candidates and selects those which are most well-suited
    to experiments
    """

    def __init__(self):
        """
        Placeholder for initializing agents, should include
        all state parameters that an agent needs to make
        decisions
        """
        pass

    @abc.abstractmethod
    def get_hypotheses(self, candidate_data, seed_data=None):
        """

        Returns:
            subset of candidate data which represent some
            choice e. g. for the next set of experiments

        """


class QBC:
    """
    Helper class for Uncertainty quantification using
    non-supporting regressors with Query-By-Committee
    """

    def __init__(self, n_members, training_fraction, model=None, test_full_model=True):
        """
        Args:
            n_members (int): Number of committee members or models to train
            training_fraction (float): fraction of data to use in training
                committee members
            model (sklearn.RegressorMixin): sklearn-style regressor
            test_full_model (bool): whether or not to test the full
                model

        """
        self.n_members = n_members
        self.training_fraction = training_fraction
        self.model = model if model else LinearRegression()
        self.committee_models = []
        self.trained = False
        self.test_full_model = test_full_model
        self.cv_score = np.nan
        self._X = None
        self._y = None

    def fit(self, X, y):
        """
        Fits the QBC committee member models

        Args:
            X (pandas.DataFrame, np.ndarray): input X values for fitting
            y (pandas.DataFrame, np.ndarray): output y values to regress
                or fit to

        Returns:
            None

        """
        self._X, self._y = X, y

        split_X = []
        split_y = []

        for i in range(self.n_members):
            a = np.arange(len(X))
            np.random.shuffle(a)
            indices = a[: int(self.training_fraction * len(X))]
            split_X.append(X.iloc[indices])
            split_y.append(y.iloc[indices])

        self.committee_models = []
        for i in tqdm(list(range(self.n_members))):
            scaler = StandardScaler()
            X = scaler.fit_transform(split_X[i])
            y = split_y[i]
            model = clone(self.model)
            model.fit(X, y)
            # Saving the scaler and model to make predictions
            self.committee_models.append([scaler, model])

        self.trained = True

        if self.test_full_model:
            # Get a CV score for an overall model with plot_hull dataset
            full_scaler = StandardScaler()
            _X = full_scaler.fit_transform(self._X, self._y)
            full_model = clone(self.model)
            full_model.fit(_X, self._y)
            cv_score = cross_val_score(
                full_model,
                _X,
                self._y,
                cv=KFold(5, shuffle=True),
                scoring="neg_mean_absolute_error",
            )
            self.cv_score = np.mean(cv_score) * -1

    def predict(self, X):
        """
        Apply the fitted committee of models to candidate space

        Args:
            X (pandas.DataFrame, np.ndarray): input matrix or values
                on which to predict

        Returns:
            (np.ndarray): mean values for predictions for all committee members
            (np.ndarray): standard deviation values for predictions for all committee members

        """
        committee_predictions = []
        for scaler, model in tqdm(self.committee_models):
            _X = scaler.transform(X)
            committee_predictions.append(model.predict(_X))
        stds = np.std(np.array(committee_predictions), axis=0)
        means = np.mean(np.array(committee_predictions), axis=0)
        return means, stds


class RandomAgent(HypothesisAgent):
    """
    Baseline agent: Randomly picks from candidate dataset
    """

    def __init__(self, candidate_data=None, seed_data=None, n_query=1):
        """
        Initializes a random agent

        Args:
            candidate_data (pandas.DataFrame): dataframe of candidates
            seed_data (pandas.DataFrame): seed data, in this case does nothing
            n_query (int): number of candidates to query

        """

        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.n_query = n_query
        super(RandomAgent, self).__init__()

    def get_hypotheses(self, candidate_data, seed_data=None):
        """

        Args:
            candidate_data (DataFrame): candidate data
            seed_data (DataFrame): seed data, there's none in this
                case, but keep the kwarg for adherence to the
                superclass signature

        Returns:
            (List) of indices

        """
        return candidate_data.sample(self.n_query)


class SequentialAgent(HypothesisAgent):
    """
    Agent that applies multiple agents sequentially, i.e.
    runs get_hypotheses multiple times to filter
    """

    def __init__(
        self,
        agents: list,
        candidate_data=None,
        seed_data=None,
    ):

        """
        Args:
            candidate_data (pandas.DataFrame): data about the candidates to search over.
                Must have a "target" column, and at least one additional column that
                can be used as descriptors.
            seed_data (pandas.DataFrame):  data which to fit the Agent to.
            agents (list): list of agents.
        """
        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.agents = agents

        super(SequentialAgent).__init__()

    def get_hypotheses(self, candidate_data, seed_data=None):
        """
        Methods for getting hypotheses using linear regression

        Args:
            candidate_data (pandas.DataFrame): candidate data
            seed_data (pandas.DataFrame): seed data

        Returns:
            (pandas.DataFrame): selected hypotheses

        """
        temp_candidate_data = candidate_data
        for agent in self.agents:
            temp_candidate_data = agent.get_hypotheses(temp_candidate_data, seed_data)
        return temp_candidate_data



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.
"""
This module contains basic campaign functionality.  Objects
and logic in this module should be very generic and not
constrained to a particular mode of materials discovery.

Furthermore, the "Campaign" logic should be kept as simple
as possible.
"""

import os
import pickle
import json
import numpy as np
import pandas as pd
import shutil
import logging

from monty.json import MSONable
from camd.utils.data import s3_sync
from camd import CAMD_S3_BUCKET
from camd.agent.base import RandomAgent


class Campaign(MSONable):
    """
    Campaign provides a sequential, workflow-like capability where an
    Agent iterates over a candidate space to choose and execute
    new Experiments, given a certain objective. The abstraction
    follows closely the "scientific method". Agent is the entity
    that suggests new Experiments.

    Supporting entities are Analyzers. Framework
    is flexible enough to implement many sequential learning or
    optimization tasks, including active-learning, bayesian optimization
    or black-box optimization with local or global optima search.

    """
    def __init__(
        self,
        candidate_data,
        agent,
        experiment,
        analyzer=None,
        seed_data=None,
        create_seed=False,
        heuristic_stopper=np.inf,
        s3_prefix=None,
        s3_bucket=CAMD_S3_BUCKET,
        path=None,
        logger=None
    ):
        """
        Invokes a campaign from candidates, seed, agent, and other supporting
        entities for decision-making, experiment, and analysis.

        Args:
            candidate_data (pd.DataFrame): List of uids for candidate
                search space for active learning
            agent (HypothesisAgent): a subclass of HypothesisAgent
            experiment (Experiment): a subclass of Experiment
            analyzer (Analyzer): a subclass of Analyzer or a list of analyzers
            seed_data (pandas.DataFrame): Seed Data for active learning,
                index is to be the assumed uid
            create_seed (int): an initial seed size to create from the data
            heuristic_stopper (int): If int, the heuristic stopper will
                kick in to check if loop should be terminated after this
                many iterations, if no discoveries in past #n loops.
            s3_prefix (str): prefix which to prepend all s3 synced files with,
                if None is specified, s3 syncing will not occur
            s3_bucket (str): bucket name for s3 sync.  If not specified,
                CAMD will sync to the specified environment variable.
            path (str): local path in which to execute the loop, defaults
                to current folder if path is not provided
            logger (logging.Logger): logger for campaign progress
        """
        # Cloud parameters
        self.s3_prefix = s3_prefix
        self.s3_bucket = s3_bucket

        # Data parameters
        self.candidate_data = candidate_data
        self.seed_data = seed_data if seed_data is not None else pd.DataFrame()
        self.create_seed = create_seed
        self.history = pd.DataFrame()

        # Object parameters
        self.agent = agent
        self.experiment = experiment
        if analyzer is not None:
            self.analyzer = analyzer if isinstance(analyzer, list) else [analyzer]
        else:
            self.analyzer = None

        # Other parameters
        # TODO: think about how to abstract this away from the loop
        self.heuristic_stopper = heuristic_stopper
        self.path = path if path else os.getcwd()
        os.chdir(self.path)

        # Internal data
        self._exp_raw_results = None

        # Check if there exists earlier iterations
        if os.path.exists(os.path.join(self.path, "iteration.json")):
            self.load("iteration")
            self.initialized = True
        else:
            self.iteration = 0
            self.initialized = False

        if self.initialized:
            self.create_seed = False
            self.load("job_status")
            self.experiment.job_status = self.job_status
            self.load("experiment", method="pickle")
            self.load("seed_data", method="pickle")
            self.load("consumed_candidates")
            self.load("loop_state", no_exist_fail=False)
            self.initialized = True
        else:
            self.submitted_experiment_requests = []
            self.consumed_candidates = []
            self.job_status = {}
            self.initialized = False
            self.loop_state = "UNSTARTED"

        if logger is None:
            self.logger = logging.Logger("camd", level=logging.INFO)
            self.logger.addHandler(logging.StreamHandler())
        else:
            self.logger = logger

    def run(self, finalize=False):
        """
        This method applies a single iteration of the loop, and
        keeps record of everything in place.

        Each iteration consists of:
            1. Get results of requested experiments
            2. Load, Expand, Save seed_data
            3. Augment candidate_space
            4. Analyze results - Stop / Go
            5. Hypothesize
            6. Submit new experiments
        """
        if not self.initialized:
            raise ValueError("Campaign must be initialized.")

        # Get new results
        self.logger.info("{} {} state: Getting new results".format(self.type, self.iteration))
        self.experiment.monitor()
        new_experimental_results = self.experiment.get_results()
        os.chdir(self.path)

        # Load seed_data
        self.load("seed_data", method="pickle")

        # Analyze new results
        self.logger.info("{} {} state: Analyzing results".format(self.type, self.iteration))
        summary = pd.DataFrame()
        if self.analyzer:
            for analyzer in self.analyzer:
                analysis = analyzer.analyze(self, finalize=finalize)
                summary = pd.concat([summary, analysis], axis=1)

        self.history = self.history.append(summary)
        self.history = self.history.reset_index(drop=True)
        self.save("history", method="pickle")

        # Augment summary and seed
        new_seed_data = self.seed_data.append(new_experimental_results)
        self.seed_data = new_seed_data
        self.save("seed_data", method="pickle")

        # Remove candidates from candidate space
        candidate_space = self.candidate_data.index.difference(
            self.consumed_candidates, sort=False
        ).tolist()
        self.candidate_data = self.candidate_data.loc[candidate_space]
        if len(self.candidate_data) == 0:
            self.logger.info("Candidate data exhausted.  Stopping loop.")
            return False

        # Campaign stopper if no discoveries in last few cycles.
        if self.iteration > self.heuristic_stopper:
            new_discoveries = self.history["new_discovery"][-3:].values.sum()
            if new_discoveries == 0:
                self.finalize()
                self.logger.info("Not enough new discoveries. Stopping the loop.")
                return False

        # Campaign stopper if finalization is desired but will be done
        # outside of run (e.g. auto_loop)
        if finalize:
            return False

        # Agent suggests new experiments
        self.logger.info(
            "{} {} state: Agent {} hypothesizing".format(
                self.type, self.iteration, self.agent.__class__.__name__
            )
        )
        suggested_experiments = self.agent.get_hypotheses(
            self.candidate_data, self.seed_data
        )

        # Campaign stopper if agent doesn't have anything to suggest.
        if len(suggested_experiments) == 0:
            self.finalize()
            self.logger.info("No agent suggestions. Stopping the loop.")
            return False

        # Experiments submitted
        self.logger.info("{} {} state: Running experiments".format(self.type, self.iteration))
        self.job_status = self.experiment.submit(suggested_experiments)
        self.save("job_status")

        # if hasattr(self.experiment, "save"):
        #     experiment.save(self.path)
        # else:
        #     self.save("experiment", method="pickle")

        self.consumed_candidates += suggested_experiments.index.values.tolist()
        self.logger.debug("{} {} state: consumed candidates {}".format(
            self.type, self.iteration, self.consumed_candidates))
        self.save("consumed_candidates")

        self.iteration += 1
        self.save("iteration")
        return True

    def auto_loop(self, n_iterations=10, monitor=False,
                  initialize=False, save_iterations=False):
        """
        Runs the loop repeatedly, and locally. Contains
        option for backing up the loop in enumerated
        subdirectories for each iteration.

        Args:
            n_iterations (int): Number of iterations.
            monitor (bool): Use Experiment's monitor method to
                keep track of requested experiments.
            initialize (bool): whether to initialize the loop
                before starting
            save_iterations (bool): whether or not to save
                iterations in subdirectories of the working
                directory

        """
        if initialize:
            self.initialize()
            if save_iterations:
                self.loop_backup("-1")
        while n_iterations - self.iteration >= 0:
            self.logger.info("Iteration: {}".format(self.iteration))
            if not self.run():
                break
            self.logger.info("  Waiting for next round ...")
            if monitor:
                self.logger.info("Monitoring experiments")
                self.experiment.monitor()
            if save_iterations:
                self.loop_backup(str(self.iteration - 1))

        # self.run(finalize=True)
        if monitor:
            self.logger.info("Monitoring experiments")
            self.experiment.monitor()
        self.finalize()

    def initialize(self, random_state=42):
        """
        Initializes a campaign. The primary goal of initialization is to ensure a proper seed exists. If create_seed
        is set in Campaign, it creates the seed by deploying the RandomAgent before the user-provided agent is
        deployed in the regular campaign iterations.
            random_state (int): ensures reproducible results.
        """
        if self.initialized:
            raise ValueError("Initialization may overwrite existing loop data. Exit.")
        if not self.seed_data.empty and not self.create_seed:
            self.logger.info(
                "{} {} state: Agent {} hypothesizing".format(
                    self.type, "initialization", self.agent.__class__.__name__
                )
            )
            suggested_experiments = self.agent.get_hypotheses(
                self.candidate_data, self.seed_data
            )
        elif self.create_seed:
            np.random.seed(seed=random_state)
            _agent = RandomAgent(self.candidate_data, n_query=self.create_seed)
            self.logger.info(
                "{} {} state: Agent {} hypothesizing".format(
                    self.type, "initialization", _agent.__class__.__name__
                )
            )
            suggested_experiments = _agent.get_hypotheses(self.candidate_data)
        else:
            raise ValueError(
                "No seed data available. Either supply or ask for creation."
            )
        # self.analyzer._initial_seed_indices = self.seed_data.index.tolist()

        self.logger.info("{} {} state: Running experiments".format(self.type, self.iteration))
        self.job_status = self.experiment.submit(suggested_experiments)
        self.consumed_candidates = suggested_experiments.index.values.tolist()
        self.logger.debug("{} {} state: consumed candidates {}".format(
            self.type, self.iteration, self.consumed_candidates))
        self.create_seed = False
        self.initialized = True

        self.save("job_status")
        self.save("seed_data", method="pickle")
        self.save("experiment", method="pickle")
        self.save("consumed_candidates")
        self.save("iteration")
        if self.s3_prefix:
            self.s3_sync()

    @property
    def type(self):
        """
        Convenience property for campaign type that
        gets the class name, mostly for logging

        Returns:
            (str): class name

        """
        return self.__class__.__name__

    def finalize(self):
        """
        Run finalization method for campaign
        if analyzer has finalize method

        Returns:
            None

        """
        self.logger.info("Finalizing campaign.")
        os.chdir(self.path)
        for analyzer in self.analyzer:
            if hasattr(analyzer, "finalize"):
                analyzer.finalize(self)
        if self.s3_prefix:
            self.s3_sync()

    def load(self, data_holder, method="json", no_exist_fail=True):
        """
        Method to load stored object attributes

        Args:
            data_holder (str): attribute to be loaded
            method (str): method by which to load object,
                'pickle' and 'json' are currently supported
            no_exist_fail (bool): whether to throw error
                on non-existence of data

        Returns:
            None

        """
        if method == "pickle":
            m = pickle
            mode = "rb"
        elif method == "json":
            m = json
            mode = "r"
        else:
            raise ValueError("Unknown data save method")

        file_name = os.path.join(self.path, data_holder + "." + method)
        exists = os.path.exists(file_name)

        if exists:
            with open(file_name, mode) as f:
                self.__setattr__(data_holder, m.load(f))
        else:
            if no_exist_fail:
                raise IOError("No {} file exists".format(data_holder))
            else:
                self.__setattr__(data_holder, None)

    def save(self, data_holder, custom_name=None, method="json"):
        """
        Save method for storing campaign data

        Args:
            data_holder (str): attribute to be written to file
            custom_name (str): custom filename if desired
            method (str): method option for data storage,
                'json' or 'pickle' are supported

        Returns:
            None

        """
        if custom_name:
            _path = os.path.join(self.path, custom_name)
        else:
            _path = os.path.join(self.path, data_holder + "." + method)
        if method == "pickle":
            m = pickle
            mode = "wb"
        elif method == "json":
            m = json
            mode = "w"
        else:
            raise ValueError("Unknown data save method")
        with open(_path, mode) as f:
            m.dump(self.__getattribute__(data_holder), f)

        if self.s3_prefix:
            self.s3_sync()

    def s3_sync(self):
        """
        Syncs current run to s3_prefix and bucket
        """
        s3_sync(self.s3_bucket, self.s3_prefix, self.path)

    def loop_backup(self, target_dir):
        """
        Helper method to backup finished loop iterations.

        Args:
            target_dir (str, Path): directory to back up to

        Returns:
            (None)

        """
        self.logger.info("Backing loop up to s3")
        os.mkdir(os.path.join(self.path, target_dir))
        _files = os.listdir(self.path)
        for file_name in _files:
            full_file_name = os.path.join(self.path, file_name)
            if os.path.isfile(full_file_name):
                shutil.copy(full_file_name, os.path.join(self.path, target_dir))

        if self.s3_prefix:
            self.s3_sync()



import unittest
import os
from camd.utils.data import cache_matrio_data, load_dataframe, \
    partition_intercomp, get_chemsys
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice
from camd import CAMD_CACHE


class DataCacheTest(unittest.TestCase):
    def setUp(self):
        # Remove smallest file and cache
        self.smallfile = "oqmd1.2_exp_based_entries_featurized_v2.pickle"
        self.smallfile_path = os.path.join(CAMD_CACHE, self.smallfile)
        if os.path.isfile(self.smallfile_path):
            os.remove(self.smallfile_path)

    def test_cache(self):
        cache_matrio_data(self.smallfile)
        self.assertTrue(os.path.isfile(self.smallfile_path))

        # Ensure no re-download of existing files in cache
        time_before = os.path.getmtime(self.smallfile_path)
        cache_matrio_data(self.smallfile)

        time_after = os.path.getmtime(self.smallfile_path)
        self.assertEqual(time_after, time_before)

    def test_load_dataframe(self):
        df_name = self.smallfile.rsplit('.', 1)[0]
        dataframe = load_dataframe(df_name)
        self.assertEqual(len(dataframe), 36581)


class PartitionTest(unittest.TestCase):
    def setUp(self):
        # Remove smallest file and cache
        self.smallfile = "oqmd1.2_exp_based_entries_featurized_v2.pickle"
        self.smallfile_path = os.path.join(CAMD_CACHE, self.smallfile)

    def test_partition(self):
        df_name = self.smallfile.rsplit('.', 1)[0]
        dataframe = load_dataframe(df_name)
        cand, seed = partition_intercomp(dataframe)
        self.assertEqual(len(dataframe), len(cand) + len(seed))

        cand, seed = partition_intercomp(dataframe, n_elements=1)
        self.assertEqual(len(dataframe), len(cand) + len(seed))
        self.assertGreater(len(seed), 0)


class GeneralUtilsTest(unittest.TestCase):
    def test_get_chemsys(self):
        # Test with structure
        struct = Structure.from_spacegroup('Fm-3m', Lattice.cubic(5.0), ['Ni', 'O'],
                                           [[0, 0, 0], [0.75, 0.25, 0.75]])
        self.assertEqual(get_chemsys(struct), 'Ni-O')

        # Test sorting
        test_val = get_chemsys("TiO2")
        self.assertEqual(test_val, "O-Ti")

        test_val = get_chemsys("BaNiO3")
        self.assertEqual(test_val, "Ba-Ni-O")


if __name__ == '__main__':
    unittest.main()



# Copyright Toyota Research Institute 2019
"""
Contains agents related to crystal-structure discovery
and hypothesizing
"""

import abc
import json
import os
from copy import deepcopy
from multiprocessing import cpu_count
from collections import OrderedDict

import numpy as np
from qmpy.analysis.thermodynamics.phase import Phase, PhaseData
from camd.analysis import PhaseSpaceAL, ELEMENTS
from camd.agent.base import HypothesisAgent, QBC
from camd.utils.data import filter_dataframe_by_composition, get_default_featurizer
from pymatgen.core.composition import Composition

from sklearn.gaussian_process.kernels import RBF, ConstantKernel
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import cross_val_score, KFold
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import Pipeline
from sklearn.ensemble import AdaBoostRegressor, BaggingRegressor
from sklearn.metrics import pairwise_distances

# TODO: Adaptive N_query and subsampling of candidate space
# TODO: Exploit/Explore general method
# TODO: General uncertainty handling?


class StabilityAgent(HypothesisAgent, metaclass=abc.ABCMeta):
    """
    The StabilityAgent is a mixin abstract class which contains
    initialization parameters and methods common to every agent
    which is responsible for making decisions about stability.
    """

    def __init__(
        self,
        candidate_data=None,
        seed_data=None,
        n_query=1,
        hull_distance=0.0,
        feature_labels=None,
        parallel=cpu_count(),
    ):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            feature_labels ([str]): list of columns to filter for features in ML
            parallel (bool, int): whether to use multiprocessing
                for phase stability analysis, if an int, sets the n_jobs
                parameter as well.  If a bool, sets n_jobs to cpu_count()
                if True and n_jobs to 1 if false.
        """
        super().__init__()
        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.n_query = n_query
        self.hull_distance = hull_distance
        self.pd = None
        self.parallel = parallel
        if feature_labels is not None:
            self.feature_labels = feature_labels
        else:
            self.feature_labels = get_default_featurizer().feature_labels()

        # These might be able to go into the base class
        self.cv_score = np.nan

    def get_pd(self, chemsys=None):
        """
        Refresh the phase diagram associated with the seed_data

        Args:
            chemsys (str): chemical system for which to filter
                seed data to provide partial phase diagram

        Returns:
            None
        """
        self.pd = PhaseData()
        # Filter seed data by relevant chemsys
        if chemsys:
            total_comp = Composition(chemsys.replace("-", ""))
            filtered = filter_dataframe_by_composition(self.seed_data, total_comp)
        else:
            filtered = self.seed_data

        phases = [
            Phase(
                row["Composition"],
                energy=row["delta_e"],
                per_atom=True,
                description=row_index,
            )
            for row_index, row in filtered.iterrows()
        ]
        phases.extend([Phase(el, 0.0, per_atom=True) for el in ELEMENTS])
        self.pd.add_phases(phases)
        return self.pd

    def update_data(self, candidate_data=None, seed_data=None):
        """
        Helper function to update the data according to the schema
        of the default OQMD data.  Updates the candidate_data and
        seed_data attributes, and returns the processed features
        and targets associated with the candidates and seed data.

        Args:
            candidate_data (DataFrame): new candidate dataframe
            seed_data (DataFrame): new seed dataframe

        Returns:
            (DataFrame): candidate features
            (DataFrame): seed features
            (DataFrame): seed targets

        """
        # Note: In the drop command, we're ignoring errors for
        #   brevity.  We should watch this, because we may not
        #   drop everything we intend to.
        if candidate_data is not None:
            self.candidate_data = candidate_data
            X_cand = self.get_features(self.candidate_data)
        else:
            X_cand = None
        if seed_data is not None:
            # Filter seed data with No delta_e, indicates a failure
            # TODO: probably worth considering putting this somewhere else
            self.seed_data = seed_data.dropna(subset=["delta_e"])
            X_seed = self.get_features(self.seed_data)
            y_seed = self.seed_data["delta_e"]
        else:
            X_seed, y_seed = None, None

        return X_cand, X_seed, y_seed

    def update_candidate_stabilities(self, formation_energies, sort=True, floor=-6.0):
        """
        Updates the candidate dataframe with the stabilities
        of the candidate compositions according to the requisite
        phase diagram analysis.

        Args:
            formation_energies ([float]): list of predictions for formation
                energies corresponding to candidate_data ordering
            sort (bool): whether or not to sort final list
            floor (float): a float intended to add a floor to the predicted
                formation energies

        Returns:
            (DataFrame): dataframe corresponding to self.candidate_data
        """
        # Preprocess formation energies with floor
        if floor is not None:
            formation_energies = np.array(formation_energies)
            formation_energies[formation_energies < floor] = floor

        # Update formation energy predictions
        self.candidate_data["pred_delta_e"] = formation_energies

        # Construct candidate phases
        candidate_phases = [
            Phase(
                data["Composition"],
                energy=data["pred_delta_e"],
                per_atom=True,
                description=m_id,
            )
            for m_id, data in self.candidate_data.iterrows()
        ]

        # Refresh and copy seed PD filtered by candidates
        all_comp = self.candidate_data["Composition"].sum()
        pd_ml = deepcopy(self.get_pd(all_comp))
        pd_ml.add_phases(candidate_phases)
        space_ml = PhaseSpaceAL(bounds=ELEMENTS, data=pd_ml)

        # Compute and return stabilities
        space_ml.compute_stabilities(candidate_phases, self.parallel)
        self.candidate_data["pred_stability"] = [
            phase.stability for phase in candidate_phases
        ]

        if sort:
            self.candidate_data = self.candidate_data.sort_values("pred_stability")

        return self.candidate_data

    def get_features(self, dataframe):
        """
        Gets features of a dataframe

        Returns:
            (pd.DataFrame): dataframe filtered to include only features
        """
        return dataframe[self.feature_labels]


class QBCStabilityAgent(StabilityAgent):
    """
    Agent which uses QBC to determine optimal hypotheses
    """

    def __init__(
        self,
        candidate_data=None,
        seed_data=None,
        n_query=1,
        hull_distance=0.0,
        parallel=cpu_count(),
        alpha=0.5,
        training_fraction=0.5,
        model=None,
        n_members=10,
        feature_labels=None,
    ):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            parallel (bool): whether to use multiprocessing
                for phase stability analysis
            training_fraction (float): fraction of data to use for
                training committee members
            alpha (float): weighting factor for the stdev in making
                best-case predictions of the stability
            model (sklearn-style regressor): regressor
            n_members (int): number of committee members for the qbc
        """

        super(QBCStabilityAgent, self).__init__(
            candidate_data=candidate_data,
            seed_data=seed_data,
            n_query=n_query,
            hull_distance=hull_distance,
            parallel=parallel,
            feature_labels=feature_labels,
        )

        self.alpha = alpha
        self.model = model
        self.n_members = n_members
        self.qbc = QBC(
            n_members=n_members,
            training_fraction=training_fraction,
            model=model,
        )

    def get_hypotheses(self, candidate_data, seed_data=None, retrain_committee=True):
        """
        Get hypotheses method for QBCStabilityAgent

        Args:
            candidate_data (pandas.DataFrame): dataframe of candidates
            seed_data (pandas.DataFrame): dataframe of prior data on
                which to fit GPUCB
            retrain_committee (bool): whether to retrain committee
                each time

        Returns:
            (pandas.DataFrame): top candidates from the GPUCB algorithm

        """
        X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)

        # Retrain committee if untrained or if specified
        if not self.qbc.trained or retrain_committee:
            self.qbc.fit(X_seed, y_seed)
        self.cv_score = self.qbc.cv_score

        # QBC makes predictions for Hf and uncertainty on candidate data
        preds, stds = self.qbc.predict(X_cand)
        expected = preds - stds * self.alpha

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)

        # Find the most stable ones up to n_query within hull_distance
        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
        within_hull = self.candidate_data[stability_filter]

        return within_hull.head(self.n_query)


class AgentStabilityML5(StabilityAgent):
    """
    An agent that does a certain fraction of full exploration and
    exploitation in each iteration.  It will exploit a fraction of
    N_query options (frac), and explore the rest of its budget.
    """

    def __init__(
        self,
        candidate_data=None,
        seed_data=None,
        n_query=1,
        hull_distance=0.0,
        parallel=cpu_count(),
        model=None,
        exploit_fraction=0.5,
        feature_labels=None,
    ):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            parallel (bool): whether to use multiprocessing
                for phase stability analysis
            model (sklearn-style Regressor): Regression method
            exploit_fraction (float): fraction of n_query to assign to
                exploitation hypotheses
        """
        super(AgentStabilityML5, self).__init__(
            candidate_data=candidate_data,
            seed_data=seed_data,
            n_query=n_query,
            hull_distance=hull_distance,
            parallel=parallel,
            feature_labels=feature_labels,
        )

        self.model = model or LinearRegression()
        self.exploit_fraction = exploit_fraction

    def get_hypotheses(self, candidate_data, seed_data=None):
        """
        Get hypotheses method for AgentStabilityML5

        Args:
            candidate_data (pandas.DataFrame): dataframe of candidates
            seed_data (pandas.DataFrame): dataframe of prior data on
                which to fit GPUCB

        Returns:
            (pandas.DataFrame): top candidates from the GPUCB algorithm

        """

        X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)
        steps = [("scaler", StandardScaler()), ("ML", self.model)]
        pipeline = Pipeline(steps)

        cv_score = cross_val_score(
            pipeline,
            X_seed,
            y_seed,
            cv=KFold(5, shuffle=True),
            scoring="neg_mean_absolute_error",
        )
        self.cv_score = np.mean(cv_score) * -1
        pipeline.fit(X_seed, y_seed)

        expected = pipeline.predict(X_cand)

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)

        # Filter by stability according to hull distance
        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
        within_hull = self.candidate_data[stability_filter]

        # Exploitation part:
        n_exploitation = int(self.n_query * self.exploit_fraction)
        to_compute = within_hull.head(n_exploitation).index.tolist()
        remaining = within_hull.tail(len(within_hull) - n_exploitation)
        remaining = remaining.append(self.candidate_data[~stability_filter])

        # Exploration part (pick randomly from remainder):
        n_exploration = self.n_query - n_exploitation

        if n_exploration > len(remaining):
            n_exploration = len(remaining)
        to_compute.extend(remaining.sample(n_exploration).index.tolist())
        return candidate_data.loc[to_compute]


class GaussianProcessStabilityAgent(StabilityAgent):
    """
    Simple Gaussian Process Regressor based Stability Agent
    """

    def __init__(
        self,
        candidate_data=None,
        seed_data=None,
        n_query=1,
        hull_distance=0.0,
        parallel=cpu_count(),
        alpha=0.5,
        feature_labels=None,
    ):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            parallel (bool): whether to use multiprocessing
                for phase stability analysis
            alpha (float): weighting factor for the stdev in making
                best-case predictions of the stability
        """
        super(GaussianProcessStabilityAgent, self).__init__(
            candidate_data=candidate_data,
            seed_data=seed_data,
            n_query=n_query,
            hull_distance=hull_distance,
            parallel=parallel,
            feature_labels=feature_labels,
        )
        self.multiprocessing = parallel
        self.alpha = alpha
        self.GP = GaussianProcessRegressor(
            kernel=ConstantKernel(1) * RBF(1), alpha=0.002
        )

    def get_hypotheses(self, candidate_data, seed_data=None):
        """
        Get hypotheses method for GaussianProcessStabilityAgent

        Args:
            candidate_data (pandas.DataFrame): dataframe of candidates
            seed_data (pandas.DataFrame): dataframe of prior data on
                which to fit GPUCB

        Returns:
            (pandas.DataFrame): top candidates from the GPUCB algorithm

        """

        X_cand, X_seed, y_seed = self.update_data(
            candidate_data=candidate_data, seed_data=seed_data
        )

        steps = [("scaler", StandardScaler()), ("GP", self.GP)]
        cv_pipeline = Pipeline(steps)
        self.cv_score = np.mean(
            -1.0
            * cross_val_score(
                cv_pipeline,
                X_seed,
                y_seed,
                cv=KFold(3, shuffle=True),
                scoring="neg_mean_absolute_error",
            )
        )

        steps = [("scaler", StandardScaler()), ("GP", self.GP)]
        overall_pipeline = Pipeline(steps)
        overall_pipeline.fit(X_seed, y_seed)

        # GP makes predictions for Hf and uncertainty*alpha on candidate data
        preds, stds = overall_pipeline.predict(X_cand, return_std=True)
        expected = preds - stds * self.alpha

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)

        # Find the most stable ones up to n_query within hull_distance
        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
        within_hull = self.candidate_data[stability_filter]

        return within_hull.head(self.n_query)


class BaggedGaussianProcessStabilityAgent(StabilityAgent):
    """
    An ensemble GP learner that can handle relatively large
    datasets by bagging. WIP.

    Current strategy is that weak GP learners are trained on
    random subsets of data with max 500 points (and no bootstrapping),
    to learn a stronger, ensemble learner that minimizes model variance.

    Learned model is usually stronger, and on par with Random Forest
    or NN. We assume the uncertainty of the prediction to be
    the minimum std among all GP estimators trained, for a given point.

    """

    def __init__(
        self,
        candidate_data=None,
        seed_data=None,
        n_query=1,
        hull_distance=0.0,
        parallel=cpu_count(),
        alpha=0.5,
        n_estimators=8,
        max_samples=5000,
        bootstrap=False,
        feature_labels=None,
    ):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            parallel (bool): whether to use multiprocessing
                for phase stability analysis
            alpha (float): scaling for stdev for estimation of best-
                case formation energies
        """
        super(BaggedGaussianProcessStabilityAgent, self).__init__(
            candidate_data=candidate_data,
            seed_data=seed_data,
            n_query=n_query,
            hull_distance=hull_distance,
            parallel=parallel,
            feature_labels=feature_labels,
        )

        self.alpha = alpha
        self.n_estimators = n_estimators
        self.max_samples = max_samples
        self.bootstrap = bootstrap

        self.GP = GaussianProcessRegressor(
            kernel=ConstantKernel(1) * RBF(1), alpha=0.002
        )

    def get_hypotheses(self, candidate_data, seed_data=None):
        """
        Get hypotheses method for BaggedGaussianProcessStabilityAgent

        Args:
            candidate_data (pandas.DataFrame): dataframe of candidates
            seed_data (pandas.DataFrame): dataframe of prior data on
                which to fit model

        Returns:
            (pandas.DataFrame): top candidates from the algorithm

        """

        X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)

        steps = [("scaler", StandardScaler()), ("GP", self.GP)]
        pipeline = Pipeline(steps)
        n_jobs = self.parallel if self.parallel else None
        bag_reg = BaggingRegressor(
            base_estimator=pipeline,
            n_estimators=self.n_estimators,
            max_samples=self.max_samples,
            bootstrap=self.bootstrap,
            verbose=True,
            n_jobs=n_jobs,
        )
        self.cv_score = np.mean(
            -1.0
            * cross_val_score(
                pipeline,
                X_seed,
                y_seed,
                cv=KFold(3, shuffle=True),
                scoring="neg_mean_absolute_error",
            )
        )
        bag_reg.fit(X_seed, y_seed)

        # GP makes predictions for Hf and uncertainty*alpha on candidate data
        preds, stds = self._get_unc(bag_reg, X_cand)
        expected = preds - stds * self.alpha

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)

        # Find the most stable ones up to n_query within hull_distance
        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
        within_hull = self.candidate_data[stability_filter]
        return within_hull.head(self.n_query)

    @staticmethod
    def _get_unc(bagging_regressor, X_test):
        """

        Args:
            bagging_regressor (RegressorMixin): regressor for which
                to get uncertainty
            X_test (np.ndarray): test data on which to estimate
                uncertainty

        Returns:
            (np.ndarray): array of uncertainty values

        """
        stds = []
        pres = []
        for est in bagging_regressor.estimators_:
            _p, _s = est.predict(X_test, return_std=True)
            stds.append(_s)
            pres.append(_p)
        return np.mean(np.array(pres), axis=0), np.min(np.array(stds), axis=0)


class AgentStabilityAdaBoost(StabilityAgent):
    """
    An agent that does a certain fraction of full exploration and
    exploitation in each iteration.  It will exploit a fraction
    of N_query options (frac), and explore the rest of its budget.
    """

    def __init__(
        self,
        candidate_data=None,
        seed_data=None,
        n_query=1,
        hull_distance=0.0,
        parallel=cpu_count(),
        model=None,
        uncertainty=True,
        alpha=0.5,
        n_estimators=10,
        exploit_fraction=0.5,
        diversify=False,
        dynamic_alpha=False,
        feature_labels=None,
    ):
        """
        Args:
            candidate_data (DataFrame): data about the candidates
            seed_data (DataFrame): data which to fit the Agent to
            n_query (int): number of hypotheses to generate
            hull_distance (float): hull distance as a criteria for
                which to deem a given material as "stable"
            parallel (bool): whether to use multiprocessing
                for phase stability analysis
            model (sklearn-style regressor): Regression method
            uncertainty (bool): whether uncertainty is included in
                minimal predictions
            alpha (float): weighting factor for the stdev in making
                best-case predictions of the stability
            n_estimators (int): number of estimators for the AdaBoosting
                algorithm
            exploit_fraction (float): fraction of n_query to assign to
                exploitation hypotheses
            diversify (bool): Turns on the diversification algorithm for
                selected experiments.
            dynamic_alpha (bool): Turns on a simple linear schedule where the
                alpha (how much uncertainty is mixed into predictions) starts
                from zero, and increases at a rate of 0.1/iter until it's capped
                at parameter alpha specified for the rest of the iterations.
        """

        super(AgentStabilityAdaBoost, self).__init__(
            candidate_data=candidate_data,
            seed_data=seed_data,
            n_query=n_query,
            hull_distance=hull_distance,
            parallel=parallel,
            feature_labels=feature_labels,
        )
        self.model = model
        self.exploit_fraction = exploit_fraction
        self.uncertainty = uncertainty
        self.alpha = alpha
        self.n_estimators = n_estimators
        self.diversify = diversify
        self.dynamic_alpha = dynamic_alpha

    def get_hypotheses(self, candidate_data, seed_data=None):
        """
        Get hypotheses method for AgentStabilityAdaBoost

        Args:
            candidate_data (pandas.DataFrame): dataframe of candidates
            seed_data (pandas.DataFrame): dataframe of prior data on
                which to fit model

        Returns:
            (pandas.DataFrame): top candidates from the algorithm

        """
        X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)

        adaboost = AdaBoostRegressor(
            base_estimator=self.model, n_estimators=self.n_estimators
        )
        steps = [("scaler", StandardScaler()), ("ML", adaboost)]
        pipeline = Pipeline(steps)

        cv_score = cross_val_score(
            pipeline,
            X_seed,
            y_seed,
            cv=KFold(3, shuffle=True),
            scoring="neg_mean_absolute_error",
        )
        self.cv_score = np.mean(cv_score) * -1

        # We will take standard scaler out of the pipleine for
        # prediction purposes (we want a single scaler)

        scaler = StandardScaler()
        X_scaled = scaler.fit_transform(X_seed)
        overall_adaboost = AdaBoostRegressor(
            base_estimator=self.model, n_estimators=self.n_estimators
        )
        overall_adaboost.fit(X_scaled, y_seed)

        X_cand = scaler.transform(X_cand)
        expected = overall_adaboost.predict(X_cand)

        if self.uncertainty:
            if self.dynamic_alpha and os.path.exists("iteration.json"):
                with open("iteration.json", "r") as f:
                    iter = json.load(f)
                print("dynamic alpha activated")
                expected -= min(0.1 * iter, self.alpha) * self._get_unc_ada(
                    overall_adaboost, X_cand
                )
            else:
                expected -= self.alpha * self._get_unc_ada(overall_adaboost, X_cand)

        # Update candidate data dataframe with predictions
        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)

        # Filter by stability according to hull distance
        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
        within_hull = self.candidate_data[stability_filter]

        # Exploitation part:
        n_exploitation = int(self.n_query * self.exploit_fraction)
        if self.diversify:
            to_compute = diverse_quant(
                within_hull.index.tolist(),
                n_exploitation,
                self.candidate_data,
                feature_filter=self.feature_labels,
            )
        else:
            to_compute = within_hull.head(n_exploitation).index.tolist()
        remaining = within_hull.tail(len(within_hull) - n_exploitation)
        remaining = remaining.append(self.candidate_data[~stability_filter])

        # Exploration part (pick randomly from remainder):
        n_exploration = self.n_query - n_exploitation

        if n_exploration > len(remaining):
            n_exploration = len(remaining)
        to_compute.extend(remaining.sample(n_exploration).index.tolist())

        return candidate_data.loc[to_compute]

    @staticmethod
    def _get_unc_ada(ada, X):
        preds = []
        for i in ada.estimators_:
            preds.append(i.predict(X))
        preds = np.array(preds)
        preds = preds.T
        stds = []
        for i in preds:
            average = np.average(i, weights=ada.estimator_weights_)
            _std = np.sqrt(
                np.average((i - average) ** 2, weights=ada.estimator_weights_)
            )
            stds.append(_std)
        return np.array(stds)


def diverse_quant(points, target_length, df, quantiles=None, feature_filter=None):
    """
    Diversify a sublist by eliminating entries based on comparisons
    with quantiles threshold and Euclidean distance.

    This method takes the points list (which would be the object
    within_hull in stability implementations), and selects a diverse
    subset for the number of exploitation choices (target_length).

    It follows a simple algorithm: start from i = 0 of the points,
    and go down the remainder of the list, removing entries that seem
    to be closer to i below a certain distance threshold. It repeats
    this process for all i: i = 1, i = 2 ... i_max.

    The method tries to adjust the distance threshold until it finds
    the shortest resulting list that is longer than target_length. If
    it can't, it will return points as it is. The threshold values are
    decided by finding distances corresponding to quantiles of the a
    sampled distribution of distances in the overall feature set.

    The method does not alter the original ordering in the list points.

    The intuition behind the algorithm is to make risk-averse choices
    by avoiding the acquisition of too similar candidates, in case one
    example among those entries is sufficient for the model to minimize
    its uncertainty and/or make a decision to not acquire any other in
    that region.

    So the resources would not be wasted and can be allocated to other
    promising choices in points.

    Args:
        points (list): Initial set of points reflecting a preferred order
        target_length (int): length of desired sublist
        df (DataFrame): feature vectors of points, where index labels
            contain elements of points
        quantiles (list): quantilies to test for threshold.
            Defaults to [0.01, 0.02, 0.03, 0.04, 0.05]

    Returns:
        A diversified sublist of points

    """
    quantiles = quantiles if quantiles else [0.01, 0.02, 0.03, 0.04, 0.05]
    if target_length >= len(points):
        return points
    if len(df) > 6000:
        _df = df.sample(6000)
    else:
        _df = df
    if feature_filter is not None:
        _df = _df[feature_filter]
    scaler = StandardScaler()
    X = scaler.fit_transform(_df)
    flatD = pairwise_distances(X).flatten()

    _df2 = df.loc[points]
    if feature_filter is not None:
        _df2 = _df2[feature_filter]
    X = scaler.transform(_df2)
    D = pairwise_distances(X)

    remove_len = len(points) - target_length
    res = []
    for alpha in quantiles:
        q = np.quantile(flatD, alpha)
        to_remove = []
        for i in range(0, len(points)):
            for j in range(i + 1, len(points)):
                if D[i, j] < q:
                    to_remove.append(j)
        _rl = len(set(to_remove))
        print(_rl, remove_len)
        if _rl >= remove_len:
            break
        res.append(to_remove)
    if len(res) == 0:
        return points[:target_length]
    else:
        d = OrderedDict()
        for i in res[-1]:  # fall back to the latest remove list before break
            d[i] = None
        final_remove_list = list(d.keys())

        return [
            p
            for p in points
            if np.where(_df2.index == p)[0][0] not in final_remove_list
        ][:target_length]



import unittest
from sklearn.model_selection import train_test_split
from camd.agent.stability import (
    QBCStabilityAgent,
    AgentStabilityML5,
    GaussianProcessStabilityAgent,
    BaggedGaussianProcessStabilityAgent,
    AgentStabilityAdaBoost,
)
from camd.utils.data import load_default_atf_data


class StabilityAgentsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        test_data = load_default_atf_data().sample(200)
        cls.seed_data, cls.candidate_data = train_test_split(
            test_data, train_size=0.2, random_state=42
        )

    def test_qbc_stability_agent(self):
        agent = QBCStabilityAgent()
        hypotheses = agent.get_hypotheses(
            candidate_data=self.candidate_data, seed_data=self.seed_data
        )

    def test_ml_agent(self):
        agent = AgentStabilityML5()
        hypotheses = agent.get_hypotheses(
            candidate_data=self.candidate_data, seed_data=self.seed_data
        )

    def test_gp_stability_agent(self):
        agent = GaussianProcessStabilityAgent()
        hypotheses = agent.get_hypotheses(
            candidate_data=self.candidate_data, seed_data=self.seed_data
        )

    def test_bagged_gp_stability_agent(self):
        agent = BaggedGaussianProcessStabilityAgent(max_samples=30)
        hypotheses = agent.get_hypotheses(
            candidate_data=self.candidate_data, seed_data=self.seed_data
        )

    def test_agent_stability_adaboost(self):
        agent = AgentStabilityAdaBoost()
        hypotheses = agent.get_hypotheses(
            candidate_data=self.candidate_data, seed_data=self.seed_data
        )


if __name__ == "__main__":
    unittest.main()



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.
"""
This module provides resources for agent optimization campaigns
"""
from camd.experiment.base import Experiment, ATFSampler
from camd.campaigns.base import Campaign
from monty.os import cd
import os


class LocalAgentSimulation(Experiment):
    """
    Class that runs Agent simulations synchronously and
    sequentially for testing in meta-agent campaigns.
    """

    def __init__(
        self,
        atf_candidate_data,
        seed_data,
        analyzer,
        iterations,
        current_data=None,
        job_status=None,
    ):
        """
        Args:
            atf_candidate_data (DataFrame): dataframe corresponding to after
                the fact data to sample
            seed_data (DataFrame): seed data to use for the campaign
            analyzer (Analyzer): Analyzer to use in the loop
            iterations (int): number of iterations to execute
                in the loop
            current_data (dataframe): current data (for restarting)
            job_status (str): job status (for restarting)

        """
        self.atf_dataframe = atf_candidate_data
        self.iterations = iterations
        self.analyzer = analyzer
        self.seed_data = seed_data
        super(LocalAgentSimulation, self).__init__(
            current_data=current_data, job_status=job_status
        )

    def submit(self, data):
        """
        Args:
            data (DataFrame): data associated with agent

        Returns:
            None

        """
        self.update_current_data(data)
        self.job_status = "PENDING"

    def monitor(self):
        """
        The monitor method in the case just runs the necessary
        agent simulation for all specified agents

        Returns:
            None

        """
        if self.job_status == "PENDING":
            campaigns = []
            for index, row in self.current_data.iterrows():
                agent = row.pop("agent")
                path = str(index)
                os.mkdir(path)
                with cd(path):
                    campaigns.append(self.test_agent(agent))
            self.current_data["campaign"] = campaigns
            self.job_status = "COMPLETED"

    def test_agent(self, agent):
        """
        Runs a simulation of a given agent according to the
        class attributes

        Args:
            agent (HypothesisAgent):

        Returns:
            None

        """
        campaign = Campaign(
            candidate_data=self.atf_dataframe,
            seed_data=self.seed_data,
            agent=agent,
            analyzer=self.analyzer,
            experiment=ATFSampler(dataframe=self.atf_dataframe),
        )
        campaign.auto_loop(n_iterations=self.iterations, initialize=True)
        return campaign

    def get_results(self):
        """
        Gets current data corresponding to last run campaign

        Returns:
            (pandas.DataFrame) current data attribute

        """
        return self.current_data



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.
"""
This module provides resources for agent optimization campaigns
"""
import pandas as pd
from taburu.table import ParameterTable
from camd import CAMD_S3_BUCKET
from camd.agent.meta import AGENT_PARAMS, convert_parameter_table_to_dataframe
from camd.agent.base import RandomAgent
from camd.campaigns.base import Campaign
from camd.analysis import AnalyzerBase
from matplotlib import pyplot as plt
import pickle
import boto3
import botocore

META_AGENT_PREFIX = "meta_agent"


class MetaAgentCampaign(Campaign):
    """
    The MetaAgentCampaign is a preliminary campaign
    which uses an array of parameters for agents
    to test agent performance in an iterative way.

    The goal of the MetaAgentCampaign is to conduct
    sequential learning on agents themselves in various
    scenarios as to determine pareto-optimally performing
    agents with various cost metrics.

    """
    @staticmethod
    def reserve(name, experiment, analyzer, agent_pool=None, bucket=CAMD_S3_BUCKET):
        """
        Method for making an reservation for a given campaign in s3

        Args:
            name (str): reservation name
            experiment (Experiment, DataFrame): complete
                experiment provisioned for agent testing,
                typically a LocalAgentSimulation
            analyzer (camd.analysis.Analyzer): analyzer for which
                to analyze/judge meta-agent campaigns
            agent_pool (ParameterTable): parameter table corresponding
                to agents to be tested
            bucket (str): name of s3 bucket

        Returns:
            (None)

        """
        # Check reservation
        client = boto3.client("s3")
        prefix = "{}/{}".format(META_AGENT_PREFIX, name)
        result = client.list_objects(Bucket=bucket, Prefix=prefix)
        if result.get("Contents"):
            raise ValueError("{} exists in s3, please choose another name".format(name))

        # Store agent pool
        agent_pool = agent_pool or ParameterTable(AGENT_PARAMS)
        pickled_agent = pickle.dumps(agent_pool)
        client.put_object(
            Bucket=bucket, Key="{}/agent_pool.pickle".format(prefix), Body=pickled_agent
        )

        # Store experiment
        pickled_experiment = pickle.dumps(experiment)
        client.put_object(
            Bucket=bucket,
            Key="{}/experiment.pickle".format(prefix),
            Body=pickled_experiment,
        )

        # Store analyzer
        pickled_analyzer = pickle.dumps(analyzer)
        client.put_object(
            Bucket=bucket,
            Key="{}/analyzer.pickle".format(prefix),
            Body=pickled_analyzer,
        )

    @staticmethod
    def update_agent_pool(name, params, bucket=CAMD_S3_BUCKET):
        """
            Function to update the agent pool associated with a campaign

            Args:
                name (str): name of campaign
                params ([{}]): list of dicts associated with agent
                    configuration parameters
                bucket (str): name of bucket to update

            Returns:
                None

            """
        client = boto3.client("s3")
        prefix = "{}/{}".format(META_AGENT_PREFIX, name)
        agent_pool = MetaAgentCampaign.load_pickled_objects(name, bucket)[0]
        agent_pool.extend(params)

        pickled_agent = pickle.dumps(agent_pool)
        client.put_object(
            Bucket=bucket, Key="{}/agent_pool.pickle".format(prefix), Body=pickled_agent
        )

    @staticmethod
    def load_pickled_objects(name, bucket=CAMD_S3_BUCKET):
        """
        Loads pickled objects for given named campaign
        and bucket

        Args:
            name (str): name of campaign
            bucket (str): name of bucket

        Returns:
            (ParameterTable): parameter table of agents
            (DataFrame): dataframe for campaign
            (Analyzer): analyzer for campaign simulated
                experiments

        """
        client = boto3.client("s3")
        prefix = "{}/{}".format(META_AGENT_PREFIX, name)
        all_objs = []
        for obj_name in ["agent_pool", "experiment", "analyzer"]:
            try:
                raw_obj = client.get_object(
                    Bucket=bucket, Key="{}/{}.pickle".format(prefix, obj_name),
                )["Body"]
            except botocore.exceptions.ClientError:
                raise ValueError(
                    "{} does not exist in s3, cannot update agent pool".format(name)
                )
            obj = pickle.loads(raw_obj.read())
            all_objs.append(obj)
        return all_objs

    @classmethod
    def from_reserved_name(cls, name, meta_agent=None, bucket=CAMD_S3_BUCKET):
        """
        Invokes a MetaAgent Campaign from a reserved name

        Args:
            name (str): name of the campaign to be run
            meta_agent (HypothesisAgent): meta-agent with
                which to select the agent simulations to
                be run, defaults to a random agent with
                n_query of 1
            bucket (str): name of the bucket from which
                to invoke the campaign

        Returns:
            (MetaAgentCampaign) - meta-agent campaign corresponding
                to the campaign

        """
        agent_pool, experiment, analyzer = cls.load_pickled_objects(name, bucket)
        s3_prefix = "{}/{}".format(META_AGENT_PREFIX, name)
        candidate_data = convert_parameter_table_to_dataframe(agent_pool)
        meta_agent = meta_agent or RandomAgent(n_query=1)
        return cls(
            candidate_data=candidate_data,
            agent=meta_agent,
            experiment=experiment,
            analyzer=analyzer,
            s3_prefix=s3_prefix,
            s3_bucket=bucket,
            create_seed=1,
        )

    def autorun(self):
        """
        Convenience method for standard running procedure

        Returns:
            (None)

        """
        self.auto_loop(n_iterations=5, initialize=True)


# TODO: move this into analysis
# For now this is specific to stability
class StabilityCampaignAnalyzer(AnalyzerBase):
    """
    The StabilityCampaignAnalyzer is an preliminary
    analyzer intended for user in meta-agent, or
    simulated discovery campaigns.  It should
    aggregate results of a campaign conducted with
    a specific agent.

    """
    def __init__(self, checkpoint_indices):
        """
        Invokes a StabilityCampaignAnalyzer using indices
        which to checkpoint.

        Args:
            checkpoint_indices ([int]): list of indices
                on which to checkpoint campaign success,
                i. e. rate of discovery
        """
        self.checkpoint_indices = checkpoint_indices

    def analyze(self, campaign, finalize=False):
        """
        Beta method for analyzing campaign results

        Args:
            campaign (Campaign): campaign object to
                be analyzed at the end of an iteration

        Returns:
            summary (DataFrame): dataframe to be appended to "history"
                that summarizes results

        """
        new_experimental_results = campaign.experiment.get_results()
        for key, row in new_experimental_results.iterrows():
            history = row.campaign.history
            for index in self.checkpoint_indices:
                new_experimental_results.loc[
                    key, "discovered_{}".format(index)
                ] = history.loc[index, "total_discovery"]
        summary = new_experimental_results.loc[
            "discovered_{}".format(self.checkpoint_indices[0]): "discovered_{}".format(
                self.checkpoint_indices[-1]
            )
        ]

        return summary

    def _plot(self, campaign_data):
        fig = plt.figure(figsize=(10, 5))
        ax1 = fig.add_subplot(1, 2, 1)
        ax1.set_title("Total")
        ax2 = fig.add_subplot(1, 2, 2)
        ax2.set_title("Cumulative")
        for idx, row in campaign_data.iterrows():
            label = "{}-{}".format(row.agent.__class__.__name__, idx)
            ax1.plot(row.campaign.history.total_discovery, label=label)
            ax2.plot(row.campaign.history.new_discovery.cumsum(), label=label)
        ax1.legend(loc="upper left")
        fig.savefig("campaign_summary.png")

    def finalize(self):
        """
        Quick method for plotting final results

        Returns:
            None

        """
        # load seed data
        df = pd.read_pickle("seed_data.pickle")
        self._plot(df)



import unittest
import pandas as pd
import os
import boto3
import json
import numpy as np

from monty.tempfile import ScratchDir
from sklearn.svm import SVR
from sklearn import preprocessing

from camd.agent.multi_fidelity import EpsilonGreedyMultiAgent, GPMultiAgent
from camd.experiment.base import ATFSampler
from camd.analysis import MultiAnalyzer
from camd.campaigns.base import Campaign
from camd import CAMD_TEST_FILES


class MultiFidelityInTandemCampaignTest(unittest.TestCase):
    def setUp(self):
        self.data = pd.read_csv(os.path.join(CAMD_TEST_FILES, "test_multi_fidelity_df.csv"),
                                index_col=0)
        self.features = [column for column in self.data if column.startswith("MagpieData")]
        self.seed_data = self.data.head(100)
        self.candidate_data = self.data.drop(self.seed_data.index)

    def test_epsilongreedy_agent(self):
        agent = EpsilonGreedyMultiAgent(candidate_data=self.candidate_data, seed_data=self.seed_data,
                                        features=self.features, target_prop='bandgap', target_prop_val=1.8,
                                        model=SVR(C=10), total_budget=10, highFi_query_frac=0.5)
        hypotheses = agent.get_hypotheses(self.candidate_data, self.seed_data)
        self.assertEqual(type(hypotheses), pd.core.frame.DataFrame)
        self.assertEqual(hypotheses.loc[hypotheses.expt_data==1].shape[0], 5)
        self.assertEqual(hypotheses.shape[0], 10)

    def test_GPLCB_agent(self):
        GP_agent = GPMultiAgent(candidate_data=self.candidate_data, seed_data=self.seed_data,
                             features=self.features, target_prop='bandgap', target_prop_val=1.8,
                             alpha=0.08, unc_percentile=100, total_budget=10)

        hypotheses = GP_agent.get_hypotheses(self.candidate_data, self.seed_data)
        self.assertEqual(type(hypotheses), pd.core.frame.DataFrame)
        self.assertEqual(hypotheses.shape[0], 10)

    def test_analyzer(self):
        sample_candidate_data = self.candidate_data.loc[[199, # good expt
                                                         237, # good expt
                                                         101, # bad expt
                                                         286, # good theory
                                                         296  # bad theory
                                                         ]]
        analyzer = MultiAnalyzer(target_prop='bandgap', prop_range=[1.6, 2.0])
        summary = analyzer._analyze_results(
            new_experimental_results=sample_candidate_data, seed_data=self.seed_data)
        self.assertEqual(
                         (summary['expt_queried'].values[0],
                          summary['total_expt_discovery'].values[0],
                          summary['new_discovery'].values[0],
                          summary['success_rate'].values[0]),
                         (3, 2, 2, 2/3)
                        )

    def test_experiment(self):
        pass

    def test_campaign(self):
        # run the campaign with epsilon greedy agent
        agent = EpsilonGreedyMultiAgent(candidate_data=self.candidate_data, seed_data=self.seed_data,
                                        features=self.features, target_prop='bandgap', target_prop_val=1.8,
                                        model=SVR(C=10), total_budget=10, highFi_query_frac=0.5)
        experiment = ATFSampler(dataframe=self.data)
        analyzer = MultiAnalyzer(target_prop='bandgap', prop_range=[1.6, 2.0])
        with ScratchDir('.'):
            campaign = Campaign(candidate_data=self.candidate_data, seed_data=self.seed_data,
                                agent=agent, experiment=experiment, analyzer=analyzer)
            campaign.auto_loop(n_iterations=2, initialize=True)
            self.assertTrue(os.path.isfile('history.pickle'))


if __name__ == '__main__':
    unittest.main()


import unittest
import os
import numpy as np

from pymatgen.core import Structure
from pymatgen.util.testing import PymatgenTest
from camd.experiment.dft import AtomateExperiment, get_mp_formation_energy
import pandas as pd
from fireworks import LaunchPad

TEST_DIR = os.path.dirname(__file__)
ATOMATE_DFT_TESTS = np.all(
    [
        os.path.isfile(os.path.join(TEST_DIR, i))
        for i in ["db.json", "my_launchpad.yaml"]
    ]
)
SKIP_MSG = (
    "Long tests disabled, provide db.json, my_launchpad.yaml "
    "in {} to run long tests".format(TEST_DIR)
)


class AtomateTest(PymatgenTest):
    # @unittest.skipUnless(ATOMATE_DFT_TESTS, SKIP_MSG)
    # def test_get(self) -> None:
    #     good_silicon = PymatgenTest.get_structure("Si")
    #     bad_silicon = good_silicon.copy()
    #
    #     # Add another site at the same position
    #     bad_silicon.append("Si", [0.1, 0.1, 0.15])
    #     bad_silicon.append("Si", [0.1, 0.333, 0.15])
    #     self.assertEqual(len(bad_silicon), 4)
    #     data = pd.DataFrame({"structure": {"good": good_silicon, "bad": bad_silicon}})
    #     db_file = os.path.join(TEST_DIR, "db.json")
    #     lpad_file = os.path.join(TEST_DIR, "my_launchpad.yaml")
    #     lpad = LaunchPad.from_file(lpad_file)
    #     lpad.auto_load()
    #     experiment = AtomateExperiment(
    #         lpad, db_file, poll_time=30, launch_from_local=False
    #     )
    #     # experiment.submit(data)
    #     # status = experiment.monitor()
    #     results = experiment.get_results()
    #
    #     self.assertAlmostEqual(
    #         results.loc["good", "final_energy_per_atom"], -5.420645, 5
    #     )
    #     self.assertIsNone(results.loc["bad", "task_id"])
    #     self.assertEqual(experiment.job_status, "COMPLETED")

    @unittest.skipUnless(ATOMATE_DFT_TESTS, SKIP_MSG)
    def test_delta_e(self) -> None:
        fe2o3 = Structure.from_file(os.path.join(TEST_DIR, "fe2o3.cif"))  # mp-19770
        data = pd.DataFrame(
            {"structure": {"mp-19770": fe2o3}, "wf_name": "opt_mp-19770", "fw_id": 2749}
        )
        db_file = os.path.join(TEST_DIR, "db.json")
        lpad_file = os.path.join(TEST_DIR, "my_launchpad.yaml")
        lpad = LaunchPad.from_file(lpad_file)
        lpad.auto_load()
        experiment = AtomateExperiment(
            lpad, db_file, poll_time=30, launch_from_local=False
        )
        experiment.update_current_data(data)
        experiment.update_results()
        self.assertAlmostEqual(experiment.current_data["delta_e"].values[0], -1.9072, 5)

    @unittest.skipUnless(ATOMATE_DFT_TESTS, SKIP_MSG)
    def testCache(self):
        db_file = os.path.join(TEST_DIR, "db.json")
        lpad_file = os.path.join(TEST_DIR, "my_launchpad.yaml")
        lpad = LaunchPad.from_file(lpad_file)
        lpad.auto_load()
        current_data = pd.read_pickle(os.path.join(TEST_DIR, "Si_test.p"))
        experiment = AtomateExperiment(
            lpad, db_file, poll_time=30, launch_from_local=False
        )
        experiment.update_current_data(current_data)
        experiment.update_results()
        results = experiment.get_results()
        self.assertAlmostEqual(
            results.loc["good", "final_energy_per_atom"], -5.420645, 5
        )
        self.assertIsNone(results.loc["bad", "task_id"])
        self.assertEqual(experiment.job_status, "COMPLETED")

    def test_eform(self):
        # SiO2 mp-546794, Fe2O3 mp-19770
        # mp-19770
        # uncorrected e -67.4928
        # corrected e -80.6388
        total_e = -67.491217
        formula = "Fe4O6"
        potcar_symbols = ["PBE Fe_pv", "PBE O"]
        hubbards = {"Fe": 5.3, "O": 0.0}
        self.assertAlmostEqual(
            get_mp_formation_energy(total_e, formula, potcar_symbols, hubbards),
            -1.9072,
            5,
        )


if __name__ == "__main__":
    unittest.main()



# Copyright Toyota Research Institute 2019
"""
Preliminary module for determining search spaces
"""
import pandas as pd
import abc
import warnings
import itertools
import numpy as np

from protosearch.build_bulk.oqmd_interface import OqmdInterface
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.core.composition import Composition, Element
from camd.utils.data import get_default_featurizer


class DomainBase(abc.ABC):
    """
    Domains combine generation and featurization and prepare
    the search space for CAMD Loop.
    """

    @abc.abstractmethod
    def candidates(self):
        """
        Primary method for every Domain to provide candidates.

        Returns:
            (pandas.DataFrame): features for generated hypothetical
                structures. The Index of dataframe should be the
                unique ids for the structures.
        """
        pass

    @property
    @abc.abstractmethod
    def bounds(self):
        """
        Returns:
            list: names of dimensions of the search space.
        """
        pass

    @abc.abstractmethod
    def sample(self, num_samples):
        """
        Abstract method for sampling from created domain

        Args:
            num_samples:
        Returns:
        """
        pass

    @property
    def bounds_string(self):
        """
        Property representation of search space bounds
        Returns:
            (str): representation of search space bounds, e.g.
                "Ir-Fe-O" or "x1-x2-x3"
        """
        return "-".join(self.bounds)


class StructureDomain(DomainBase):
    """
    Provides machine learning ready candidate domains (search spaces) for
    hypothetical structures.  If scanning an entire system, use the
    StructureDomain.from_bounds method. If scanning for formula(s),
    provide a list of formulas directly to StructureDomain.

    Once the StructureDomain is initialized, the method candidates returns
    a fully-featurized hypothetical materials set subject to n_max_atoms.
    """

    def __init__(self, formulas, n_max_atoms=None):
        """
        Args:
            formulas ([str]): list of chemical formulas to create new
                material candidates.
            n_max_atoms (int): number of max atoms
        """
        self.formulas = formulas
        self.n_max_atoms = n_max_atoms
        self.features = None
        self._hypo_structures = None

    @classmethod
    def from_bounds(
        cls,
        bounds,
        n_max_atoms=None,
        charge_balanced=True,
        create_subsystems=False,
        **kwargs
    ):
        """
        Convenience constructor that delivers an ML-ready domain
        from defined chemical boundaries.

        Args:
            bounds ([str]): list of element strings corresponding to bounds of
                the composition space, e. g. ['Fe', 'O', 'N']
            n_max_atoms (int): maximum number of atoms in the generated
                formulae
            charge_balanced (bool): whether to filter generated formulae by
                charge balancing the respective elements according to allowed
                oxidation states
            create_subsystems (bool): TODO - what is this?
            **kwargs: arguments to pass to formula creator
        """
        formulas = create_formulas(
            bounds,
            charge_balanced=charge_balanced,
            create_subsystems=create_subsystems,
            **kwargs
        )
        print("Generated chemical formulas: {}".format(formulas))
        return cls(formulas, n_max_atoms)

    @property
    def bounds(self):
        """
        Method to get bounds from StructureDomain

        Returns:
            ([]): list of dimensions in search space

        """
        bounds = set()
        for formula in self.formulas:
            bounds = bounds.union(Composition(formula).as_dict().keys())
        return bounds

    def get_structures(self):
        """
        Method to call protosearch structure generation
        """
        if self.formulas:
            print("Generating hypothetical structures...")
            self._hypo_structures = get_structures_from_protosearch(self.formulas)
            print(
                "Generated {} hypothetical structures".format(len(self.hypo_structures))
            )
        else:
            raise ValueError("Need formulas to create structures")

    @property
    def hypo_structures(self):
        """
        Returns (dataframe): Hypothetical structures generated by
            protosearch, filtered by n_max_atoms

        """
        if self._hypo_structures is None:
            self.get_structures()
        if self.n_max_atoms:
            n_max_filter = [
                i.num_sites <= self.n_max_atoms
                for i in self._hypo_structures["structure"]
            ]
            if self._hypo_structures is not None:
                return self._hypo_structures[n_max_filter]
            else:
                return None
        else:
            return self._hypo_structures

    @property
    def hypo_structures_dict(self):
        """
        Returns:
            (dict): Hypothetical structures generated by
                protosearch, filtered by n_max_atoms

        """
        return self.hypo_structures["structure"].to_dict()

    @property
    def compositions(self):
        """
        Returns:
            (list): Compositions of hypothetical structures generated.
        """
        if self.hypo_structures is not None:
            return [s.composition for s in self.hypo_structures]
        else:
            warnings.warn("No stuctures available.")
            return []

    @property
    def formulas_with_valid_structures(self):
        """
        Quick method to filter formulas with valid structures

        Returns:
            ([str]): list of formulas with corresponding valid
                structures

        """
        # Note the redundancy here is for pandas to work
        if self.valid_structures is not None:
            return [s.composition.formula for s in self.valid_structures["structure"]]
        else:
            warnings.warn("No structures available yet.")
            return []

    def featurize_structures(self, featurizer=None, **kwargs):
        """
        Featurizes the hypothetical structures available from
        hypo_structures method. Hypothetical structures for which
        featurization fails are removed and valid structures are
        made available as valid_structures

        Args:
            featurizer (Featurizer): A MatMiner Featurizer.
                Defaults to MultipleFeaturizer with PRB Ward
                Voronoi descriptors.
            **kwargs (dict): kwargs passed to featurize_many
                method of featurizer.

        Returns:
            (pandas.DataFrame): features

        """
        # Note the redundancy here is for pandas to work
        if self.hypo_structures is None:
            warnings.warn("No structures available. Generating structures.")
            self.get_structures()

        print("Generating features")
        featurizer = (
            featurizer
            if featurizer
            else get_default_featurizer()
        )

        features = featurizer.featurize_many(
            self.hypo_structures["structure"], ignore_errors=True, **kwargs
        )

        n_species, formula = [], []
        for s in self.hypo_structures["structure"]:
            n_species.append(len(s.composition.elements))
            formula.append(s.composition.formula)

        self._features_df = pd.DataFrame.from_records(
            features, columns=featurizer.feature_labels()
        )
        self._features_df.index = self.hypo_structures.index
        self._features_df["N_species"] = n_species
        self._features_df["Composition"] = formula
        self._features_df["structure"] = self.hypo_structures["structure"]
        self.features = self._features_df.dropna(axis=0, how="any")
        self.features = self.features.reindex(sorted(self.features.columns), axis=1)

        self._valid_structure_labels = list(self.features.index)
        self.valid_structures = self.hypo_structures.loc[self._valid_structure_labels]

        print(
            "{} out of {} structures were successfully featurized.".format(
                self.features.shape[0], self._features_df.shape[0]
            )
        )
        return self.features

    def candidates(self, include_composition=True):
        """
        This is the recommended convenience method that returns
        a fully-featurized set of hypothetical structures.

        Args:
            include_composition (bool): Adds a column named "formula"
                to the dataframe.

        Returns:
            (pandas.DataFrame) feature vectors of valid
                hypothetical structures.
        """
        if self._hypo_structures is None:
            self.get_structures()

        if self.features is None:
            self.featurize_structures()
        if include_composition:
            return self.features
        else:
            return self.features.drop("Composition", axis=1)

    def sample(self, num_samples):
        """
        Method for sampling domain

        Args:
            num_samples (int): number of samples to return

        Returns:
            (pd.DataFrame): dataframe corresponding to sampled
                domain with num_samples candidates

        """
        self.candidates().sample(num_samples)


def get_structures_from_protosearch(formulas, source="icsd", db_interface=None):
    """
    Calls protosearch to get the hypothetical structures.

    Args:
        formulas ([str]): list of chemical formulas from which
            to generate candidate structures
        source (str): project name in OQMD to be used as source.
            Defaults to ICSD.
        db_interface (DbInterface): interface to OQMD database
            by default uses the one pulled from data.matr.io

    Returns:
        (pandas.DataFrame) hypothetical pymatgen structures
            generated and their unique ids from protosearch
    """

    if db_interface is None:
        db_interface = OqmdInterface(source)
    dataframes = [
        db_interface.create_proto_data_set(chemical_formula=formula)
        for formula in formulas
    ]
    _structures = pd.concat(dataframes)

    # Drop bad structures
    _structures.dropna(axis=0, how="any", inplace=True)

    # conversion to pymatgen structures
    ase_adap = AseAtomsAdaptor()
    pmg_structures = [
        ase_adap.get_structure(_structures.iloc[i]["atoms"])
        for i in range(len(_structures))
    ]
    _structures["structure"] = pmg_structures
    # This is for compatibility with Mc1, which doesn't allow
    # underscores
    structure_uids = [
        _structures.iloc[i]["structure_name"].replace('_', '-')
        for i in range(len(_structures))
    ]
    _structures.index = structure_uids
    return _structures


def get_stoichiometric_formulas(n_components, grid=None):
    """
    Generates anonymous stoichiometric formulas for a set
    of n_components with specified coefficients

    Args:
        n_components (int): number of components (dimensions)
        grid (list): a range of integers

    Returns:
        (list): unique stoichiometric formula from an
            allowed grid of integers.
    """
    grid = grid if grid else list(range(1, 8))
    args = [grid for _ in range(n_components)]
    stoics = np.array(list(itertools.product(*args)))
    fracs = stoics.astype(float) / np.sum(stoics, axis=1)[:, None]
    _, indices, counts = np.unique(fracs, axis=0, return_index=True, return_counts=True)
    return stoics[indices]


def create_formulas(
    bounds,
    charge_balanced=True,
    oxi_states_extend=None,
    oxi_states_override=None,
    all_oxi_states=False,
    grid=None,
    create_subsystems=False,
):
    """
    Creates a list of formulas given the bounds of a chemical space.

    TODO:
        - implement create_subsystems

    Args:
        bounds ([str]): list of elements to bound the space
        charge_balanced (bool): whether to balance oxidations
            states in the generated formulae
        oxi_states_extend ({}): dictionary of {element: [int]}
            where the value is the added oxidation state to be
            included
        oxi_states_override ({str: int}): override for oxidation
            states, see Composition.oxi_state_guesses
        all_oxi_states ({str: int): global config for oxidation
            states, see Composition.oxi_state_guesses
        grid ([]): list of integers to use for coefficients
        create_subsystems (bool): whether to create formulas
            for sub-chemical systems, e. g. for Sr-Ti-O,
            whether to create Ti-O and Sr-O

    Returns:
        ([str]): list of chemical formulas

    """
    if create_subsystems:
        raise NotImplementedError("Create subsystems not yet implemented.")

    stoichs = get_stoichiometric_formulas(len(bounds), grid=grid)

    formulas = []
    for f in stoichs:
        f_ = ""
        for i in range(len(f)):
            f_ += bounds[i] + f.astype(str).tolist()[i]
        formulas.append(f_)

    if charge_balanced:
        charge_balanced_formulas = []

        if oxi_states_extend:
            oxi_states_override = oxi_states_override if oxi_states_override else {}
            for element, states in oxi_states_extend.items():
                states = states if isinstance(states, list) else [states]
                _states = states + list(Element[element].common_oxidation_states)
                if element in oxi_states_override:
                    oxi_states_override[element] += states
                else:
                    oxi_states_override[element] = _states

        for formula in formulas:
            c = Composition(formula)
            if c.oxi_state_guesses(
                oxi_states_override=oxi_states_override, all_oxi_states=all_oxi_states
            ):
                charge_balanced_formulas.append(formula)
        return charge_balanced_formulas
    else:
        return formulas


def heuristic_setup(elements):
    """
    Helper function to setup a default structure_domain

    Args:
        elements ([str]): list of elements to use to
            generate formulae

    Returns:
        (int): maximum coefficient for element set
        (bool): whether or not charge balancing should be used

    """
    grid_defaults = {2: 5, 3: 5}
    n_comp = len(elements)
    _g = grid_defaults.get(n_comp, 4)

    # Charge balance ionic compounds
    if {"O", "Cl", "F", "S", "N", "Br", "I"}.intersection(set(elements)):
        charge_balanced = True
    else:
        charge_balanced = False

    if not charge_balanced:
        return _g, charge_balanced
    else:
        g_max_max = 8
        while True:
            sd = StructureDomain.from_bounds(
                elements, charge_balanced=True, grid=range(1, _g)
            )
            n = len(sd.formulas)
            if n >= 20:
                return _g, charge_balanced
            else:
                if _g < g_max_max:
                    _g += 1
                else:
                    return _g, charge_balanced



"""
This module consolidates generic aws s3 utilities
"""

import os

import boto3
import requests
import pandas as pd
import numpy as np
from pymatgen import Composition
from monty.os import makedirs_p
from camd import CAMD_CACHE, tqdm


def s3_sync(s3_bucket, s3_prefix, sync_path="."):
    """
    Syncs a given path to an s3 prefix

    Args:
        s3_bucket (str): bucket name
        s3_prefix (str): s3 prefix to sync to
        sync_path (str, Path): path to sync to bucket:prefix

    Returns:
        (None)

    """
    # Get bucket
    s3_resource = boto3.resource("s3")
    bucket = s3_resource.Bucket(s3_bucket)

    # Walk paths and subdirectories, uploading files
    for path, subdirs, files in os.walk(sync_path):
        # Get relative path prefix
        relpath = os.path.relpath(path, sync_path)
        if not relpath.startswith('.'):
            prefix = os.path.join(s3_prefix, relpath)
        else:
            prefix = s3_prefix

        for file in files:
            file_key = os.path.join(prefix, file)
            bucket.upload_file(os.path.join(path, file), file_key)



# Copyright Toyota Research Institute 2022
from camd.campaigns.base import Campaign
from sklearn.neural_network import MLPRegressor
from camd.agent.stability import AgentStabilityAdaBoost
from camd.analysis import StabilityAnalyzer
from camd.experiment.base import Experiment
from camd.experiment.dft import get_mp_formation_energy_from_m3gnet
import pandas as pd
import os
import shutil
from m3gnet.models import Relaxer
from tqdm import tqdm

import warnings
warnings.filterwarnings('ignore')


if __name__ == "__main__":
    # We'll be using a cached dataset, but here's an example of how you
    # can fetch data from the materials project if needed
    # from pymatgen.ext.matproj import MPRester
    # from camd.utils.data import get_default_featurizer
    # with MPRester() as mpr:
    #    data = mpr.query({"nelements": {"$lte": 2}, "elements": "S"},
    #              ['structure', 'material_id',
    #               'pretty_formula', 'formation_energy_per_atom'])
    # data = {r['material_id']: r for r in data}
    # dataframe = pd.DataFrame.from_dict(data, orient='index')
    # featurizer = get_default_featurizer()
    # dataframe = featurizer.featurize_dataframe(dataframe, 'structure', ignore_errors=True)
    #  dataframe = dataframe.rename(columns={
    #      "formation_energy_per_atom": "delta_e",
    #      "pretty_formula": "Composition"
    #  })
    # dataframe.to_pickle("binary_sulfides_mp.pickle")


    # In[4]:


    dataframe = pd.read_pickle("binary_sulfides_mp.pickle")
    dataframe = dataframe.dropna()


    # In[5]:


    dataframe[['Composition', 'delta_e']].head()


    # In[6]:


    # CAMD generates structure using a structure generation algorithm
    # defined by Wyckoff-site enumerated prototypes, we'll also cache
    # these because featurization takes a little while, but here's the code
    use_cached = True
    if not use_cached:
        from camd.domain import StructureDomain, heuristic_setup
        chemsys = ["Mn", "S"]
        max_coeff, charge_balanced = heuristic_setup(chemsys)
        domain = StructureDomain.from_bounds(
            chemsys,
            charge_balanced=charge_balanced,
            n_max_atoms=20,
            grid=(1, max_coeff)
        )
        candidate_data = domain.candidates()
        candidate_data.to_pickle("mn_s_candidates.pickle")
    else:
        candidate_data = pd.read_pickle("mn_s_candidates.pickle")


    # In[ ]:





    # In[7]:


    # Define the experiment
    # Experiment classes in CAMD require 3 methods
    # - submit starts the calculation
    # - get_results returns a dataframe with results
    # - monitor is a method to wait until results are done, which ensures
    #   compatibility with CAMD logic
    class M3GNetExperiment(Experiment):
        def submit(self, data):
            """data is a pandas dataframe that must have a column called structure"""
            # Sometimes in this method you might submit to an external system,
            # e.g. AWS batch or a supercomputer, but in this case we just update
            # the data because we do the computation locally
            self.update_current_data(data)

        def monitor(self):
            relaxer = Relaxer()
            delta_es = []
            for index, row in tqdm(self.current_data.iterrows()):
                s = row['structure']
                t = relaxer.relax(s)["trajectory"]
                e = t.energies[-1].flatten()[0]
                delta_e = get_mp_formation_energy_from_m3gnet(
                    e, s
                )
                delta_es.append(delta_e)
            self.current_data['delta_e'] = delta_es

        def get_results(self):
            return self.current_data

        def save(self):
            return None


    # In[8]:


    ##########################################################
    # Binary stable material discovery 50:50 explore/exploit agent
    ##########################################################
    n_query = 5  # This many new candidates are "calculated with DFT"
    # (i.e. requested from Oracle -- DFT)
    agent = AgentStabilityAdaBoost(
        model=MLPRegressor(hidden_layer_sizes=(40, 20)),
        n_query=n_query,
        hull_distance=0.2,
        uncertainty=True,
        exploit_fraction=0.8, # Number of candidates chosen by the ML model, rest are random
        alpha=0.5, # Weighting of the estimated uncertainty from the AdaBoost ensemble
        n_estimators=10
    )
    analyzer = StabilityAnalyzer(hull_distance=0.2)
    experiment = M3GNetExperiment()
    candidate_data = dataframe

    # Usually takes ~10 minutes
    path = os.path.join(os.getcwd(), "m3gnet_structure_discovery")
    shutil.rmtree(path, ignore_errors=True)
    os.makedirs(path)
    new_loop = Campaign(
        candidate_data,
        agent,
        experiment,
        analyzer,
        seed_data=dataframe,
        path=path
    )
    new_loop.auto_loop(n_iterations=6, initialize=True)

    new_loop.history.plot.bar(y='total_discovery')


import unittest

import pandas as pd
import numpy as np
from camd.experiment.base import ATFSampler


class ATFSamplerTest(unittest.TestCase):
    def test_submit_get_results(self):
        test_dataframe = pd.DataFrame(
            {"index": np.arange(5), "squared": np.arange(5) ** 2}
        )
        simple_exp = ATFSampler(test_dataframe)
        simple_exp.submit(test_dataframe.loc[[0, 2, 3]])
        self.assertEqual(simple_exp.job_status, "COMPLETED")
        simple_exp.monitor()
        self.assertTrue((simple_exp.get_results()["squared"] == [0, 4, 9]).all())


if __name__ == "__main__":
    unittest.main()



"""
Script test of MProtoDFTCampaign
"""
from datetime import datetime
import logging
from fireworks import LaunchPad
from sklearn.neural_network import MLPRegressor
from camd.domain import StructureDomain, heuristic_setup
from camd.agent.m3gnet import M3GNetStabilityAgent
from camd.agent.stability import AgentStabilityAdaBoost
from camd.agent.base import SequentialAgent
from camd.campaigns.structure_discovery import ProtoDFTCampaign
from camd.experiment.dft import AtomateExperiment
from camd.analysis import StabilityAnalyzer
import pandas as pd
from monty.os import cd, makedirs_p

from m3gnet.models import M3GNet


def load_candidate(chemsys, n_max_atoms=20):
    """
    Load candidate data
    Args:
        chemsys (str): chemical system
        n_max_atoms (int): number of maximum atoms
    """
    element_list = chemsys.split("-")
    max_coeff, charge_balanced = heuristic_setup(element_list)
    domain = StructureDomain.from_bounds(
        element_list,
        charge_balanced=charge_balanced,
        n_max_atoms=n_max_atoms,
        **{"grid": range(1, max_coeff)}
    )
    candidate_data = domain.candidates()
    return candidate_data


def load_seed():
    """
    Helper function
    Returns:
        list of the MP entries (w/ structure)
        in the chemsys
    """
    # with MPRester() as mpr:
    #     entries = mpr.get_entries_in_chemsys(chemsys, inc_structure=True)
    # return entries
    data = pd.read_pickle("mp_binary.pickle")
    data = data.rename(
        columns={
            "formation_energy_per_atom": "delta_e",
            "pretty_formula": "Composition",
        }
    )
    data = data.dropna()
    return data


if __name__ == "__main__":
    # Params for experiment
    DB_FILE = "/mnt/efs/fw_config/db.json"
    LPAD_FILE = "/mnt/efs/fw_config/my_launchpad.yaml"
    # Parameter
    CHEMSYS = "Si"
    NOW = datetime.now().isoformat()
    seed_data = load_seed()
    exp_data = load_candidate(CHEMSYS)
    model = M3GNet(is_intensive=False)
    m3gnetagent = M3GNetStabilityAgent(m3gnet=model, hull_distance=1.0, n_query=5)
    adagent = AgentStabilityAdaBoost(
        model=MLPRegressor(hidden_layer_sizes=(84, 50)),
        n_query=5,
        hull_distance=0.3,
        exploit_fraction=1.0,
        uncertainty=True,
        alpha=0.5,
        diversify=True,
        n_estimators=20,
    )
    agent = SequentialAgent(agents=[adagent, m3gnetagent])
    analyzer = StabilityAnalyzer(hull_distance=0.2)
    lpad = LaunchPad.from_file(LPAD_FILE)
    lpad.auto_load()
    experiment = AtomateExperiment(lpad, DB_FILE, poll_time=60, launch_from_local=False)

    path = "campaigns/{}".format(NOW)
    makedirs_p(path)
    logger = logging.Logger("camd")
    logger.setLevel("INFO")
    log_file = "{}/log.txt".format(path)
    file_handler = logging.FileHandler(log_file)
    logger.addHandler(file_handler)
    logger.addHandler(logging.StreamHandler())

    with cd(path):
        campaign = ProtoDFTCampaign(
            candidate_data=exp_data,
            agent=agent,
            experiment=experiment,
            analyzer=analyzer,
            seed_data=seed_data,
            heuristic_stopper=5,
            logger=logger,
        )
        campaign.autorun()
        with open("done", "w") as f:
            f.write(datetime.now().isoformat())



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.

import unittest
import boto3
import os
import pandas as pd

os.environ["CAMD_S3_BUCKET"] = "camd-test"
from camd.campaigns.structure_discovery import ProtoDFTCampaign
from camd.agent.stability import AgentStabilityML5, AgentStabilityAdaBoost
from camd.analysis import StabilityAnalyzer
from camd.experiment.base import ATFSampler
from camd import CAMD_S3_BUCKET, CAMD_TEST_FILES
from camd.utils.data import filter_dataframe_by_composition, load_dataframe
from monty.tempfile import ScratchDir
from sklearn.neural_network import MLPRegressor
from m3gnet.models import M3GNet

from camd import CAMD_TEST_FILES
from camd.agent.m3gnet import M3GNetHypothesisAgent, M3GNetStabilityAgent, reverse_calcs
from camd.agent.base import SequentialAgent

CAMD_DFT_TESTS = os.environ.get("CAMD_DFT_TESTS", False)
SKIP_MSG = "Long tests disabled, set CAMD_DFT_TESTS to run long tests"


def teardown_s3():
    """Tear down test files in s3"""
    s3 = boto3.resource("s3")
    bucket = s3.Bucket(CAMD_S3_BUCKET)
    bucket.objects.filter(Prefix="proto-dft-2/runs/Si").delete()
    bucket.objects.filter(Prefix="proto-dft-2/submit/Si").delete()


class ProtoDFTCampaignTest(unittest.TestCase):
    def test_simulated(self):
        # Note that there's a small issue with pickled results here that may not have
        # certain spin and magnetization flags set - pickled Vasprun objects may not
        # be completely compatible with latest version of pymatgen
        exp_dataframe = pd.read_pickle(
            os.path.join(CAMD_TEST_FILES, "mn-ni-o-sb.pickle")
        )
        experiment = ATFSampler(exp_dataframe)
        candidate_data = exp_dataframe.iloc[:, :-11]
        # Set up agents and loop parameters
        agent = AgentStabilityAdaBoost(
            model=MLPRegressor(hidden_layer_sizes=(40, 20)),
            n_query=2,
            hull_distance=0.2,
            exploit_fraction=1.0,
            uncertainty=True,
            alpha=0.5,
            diversify=True,
            n_estimators=5,
        )
        analyzer = StabilityAnalyzer(hull_distance=0.2)
        # Reduce seed_data
        icsd_data = load_dataframe("oqmd1.2_exp_based_entries_featurized_v2")
        seed_data = filter_dataframe_by_composition(icsd_data, "MnNiOSb")
        leftover = ~icsd_data.index.isin(seed_data.index)
        # Add some random other data to test compositional flexibility
        seed_data = seed_data.append(icsd_data.loc[leftover].sample(30))
        del icsd_data
        with ScratchDir("."):
            campaign = ProtoDFTCampaign(
                candidate_data=candidate_data,
                agent=agent,
                experiment=experiment,
                analyzer=analyzer,
                seed_data=seed_data,
                heuristic_stopper=5,
            )
            campaign.autorun()
            self.assertTrue(os.path.isfile('hull_finalized.png'))

    def test_sim_m3gnet(self):
        exp_dataframe = pd.read_pickle(
            os.path.join(CAMD_TEST_FILES, "test_m3gnet_featurized.pickle")
        )
        seed_data = pd.read_pickle(
            os.path.join(CAMD_TEST_FILES, "mp_test_data.pickle")
        )
        seed_data = seed_data.rename(
            columns={
                "formation_energy_per_atom": "delta_e",
                "pretty_formula": "Composition",
            }
        )
        seed_data = seed_data.dropna()
        seed_data.index = ["mp-{}".format(n) for n in seed_data.index]
        exp_dataframe.index = ["camd-{}".format(n) for n in exp_dataframe.index]
        experiment = ATFSampler(exp_dataframe)
        exp_dataframe["status"] = "SUCCEEDED"
        # candidate_data = exp_dataframe.iloc[:, :-11]
        # Set up agents and loop parameters
        model = M3GNet(is_intensive=False)
        m3gnetagent = M3GNetStabilityAgent(m3gnet=model, hull_distance=2.0)
        adagent = AgentStabilityAdaBoost(
            model=MLPRegressor(hidden_layer_sizes=(20,)),
            n_query=3,
            hull_distance=100.0,
            exploit_fraction=1.0,
            uncertainty=True,
            alpha=0.5,
            diversify=True,
            n_estimators=4,
        )
        agent = SequentialAgent(agents=[adagent, m3gnetagent])
        analyzer = StabilityAnalyzer(hull_distance=0.2)
        with ScratchDir("."):
            campaign = ProtoDFTCampaign(
                candidate_data=exp_dataframe.drop("delta_e", axis=1),
                agent=agent,
                experiment=experiment,
                analyzer=analyzer,
                seed_data=seed_data,
                heuristic_stopper=5,
            )
            campaign.autorun()

    @unittest.skipUnless(CAMD_DFT_TESTS, SKIP_MSG)
    def test_cached_campaign(self):
        with ScratchDir("."):
            campaign = ProtoDFTCampaign.from_chemsys_high_quality("Si")
            # Test seed data has other data
            self.assertGreater(len(campaign.seed_data), 36581)

    @unittest.skipUnless(CAMD_DFT_TESTS, SKIP_MSG)
    def test_simple_dft(self):
        with ScratchDir("."):
            campaign = ProtoDFTCampaign.from_chemsys("Si")
            # Nerf agent a bit
            agent = AgentStabilityML5(n_query=2)
            campaign.agent = agent
            campaign.autorun()
            teardown_s3()


if __name__ == "__main__":
    unittest.main()



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.
"""
Module and script for running the CAMD campaign worker

Usage:
    camd_worker COMMAND [options]

Options:
    --campaign CAMPAIGN  campaign name  [default: proto-dft-2]
    --loops NUM_LOOPS    number of loops to run
    --catch_errors       whether or not to catch errors
    -h --help            Show this screen
    --version            Show version

COMMAND may be one of:
    start - starts worker
    stop - writes stopfile such that CAMD stops
"""
import time
import boto3
import itertools
import os
import numpy as np
import traceback

from monty.serialization import dumpfn
from pathlib import Path
from docopt import docopt
from monty.tempfile import ScratchDir
from camd import CAMD_S3_BUCKET, CAMD_STOP_FILE
from camd.campaigns.structure_discovery import \
    ProtoDFTCampaign, CloudATFCampaign
from camd.campaigns.meta_agent import MetaAgentCampaign


class Worker(object):
    """
    The Worker is an object that is intended to
    persistently poll s3 for new submissions
    from which to start campaigns.  Currently
    primarily used for structure discovery,
    i.e. 'proto-dft' campaigns.
    """
    def __init__(self, campaign="proto-dft-2",
                 catch_errors=False):
        """
        Initialize a Worker

        Args:
            campaign (str): campaign name, e.g. 'proto-dft-2'

        """

        self.campaign = campaign
        self.catch_errors = catch_errors

    def start(self, num_loops=np.inf, sleep_time=60):
        """
        Starts the worker, which monitors s3 for new submissions
        and starts runs accordingly.

        Args:
            num_loops (int): number of campaigns to run before
                stopping
            sleep_time (float): time to sleep between iterations
                in which active campaigns are found

        Returns:
            (int) number of loops executed, including "sleep" loops,
                where no campaign was executed

        """
        for loop_num in itertools.count(1):
            if loop_num > num_loops or self.check_stop_file():
                return loop_num - 1
            latest_submission = self.get_latest_submission()
            if latest_submission:
                with ScratchDir('.') as sd:
                    print("Running {} in {}".format(latest_submission, sd))
                    self.run_campaign(latest_submission)
            else:
                print("No new campaigns submitted, sleeping for {} seconds".format(
                    sleep_time))
                time.sleep(sleep_time)

    def run_campaign(self, *args):
        """
        Runs the campaign for a given chemsys

        Args:
            *args: args for a given campaign,
                specific to invocation method below i. e.

                proto-dft/oqmd-atf require CHEMSYS
                meta-agent requires NAME

        Returns:
            None

        """
        if self.campaign.startswith("proto-dft-high"):
            campaign = ProtoDFTCampaign.from_chemsys_high_quality(*args)
        elif self.campaign.startswith("proto-dft"):
            campaign = ProtoDFTCampaign.from_chemsys(*args)
        elif self.campaign.startswith("oqmd-atf"):
            # This is more or less just a test
            campaign = CloudATFCampaign.from_chemsys(*args)
        elif self.campaign.startswith("meta-agent"):
            # For meta-agent campaigns, submit with meta_agent/CAMPAIGN_NAME
            campaign = MetaAgentCampaign.from_reserved_name(*args)
        else:
            raise ValueError("Campaign {} is not valid".format(self.campaign))

        # Capture errors and store
        if self.catch_errors:
            try:
                campaign.autorun()
            except Exception as e:
                error_msg = {"error": "{}".format(e),
                             "traceback": traceback.format_exc()}
                campaign.logger.info("Error: {}".format(e))
                campaign.logger.info("Traceback: {}".format(traceback.format_exc()))
                dumpfn(error_msg, "error.json")
                dumpfn({"status": "error"}, "job_status.json")
                campaign.s3_sync()
        else:
            campaign.autorun()

    def get_latest_submission(self):
        """
        Gets the last submitted chemsys

        Returns:
            (str): corresponding to last submitted chemsys

        """
        bucket = boto3.resource("s3").Bucket(CAMD_S3_BUCKET)

        # Get submissions
        submission_prefix = '/'.join([self.campaign, "submit"])
        submit_objects = bucket.objects.filter(Prefix=submission_prefix)
        submission_times = {obj.key.split('/')[-2]: obj.get()['LastModified']
                            for obj in submit_objects
                            if obj.key != "{}/submit/".format(self.campaign)}

        # Get started jobs
        started = get_common_prefixes(CAMD_S3_BUCKET, "/".join([self.campaign, "runs"]))

        # Filter started jobs and then get latest unstarted
        unstarted = list(set(submission_times.keys()) - set(started))
        latest_unstarted = sorted(unstarted, key=lambda x: submission_times[x])

        return latest_unstarted[-1] if latest_unstarted else None

    @staticmethod
    def write_stop_file():
        """
        Touches stop file to signal drawdown to shared workers

        Returns:
            (None)

        """
        Path(CAMD_STOP_FILE).touch()

    @staticmethod
    def remove_stop_file():
        """
        Removes the stop file from its specified location
        via CAMD_STOP_FILE env variable

        Returns:
            (None)

        """
        if os.path.isfile(CAMD_STOP_FILE):
            os.remove(CAMD_STOP_FILE)

    @staticmethod
    def check_stop_file():
        """
        Returns bool corresponding to whether the CAMD stop
        file exists

        Returns:
            (bool) whether CAMD stop file exists

        """
        return os.path.isfile(CAMD_STOP_FILE)


# TODO: move this to TRI-utils
def get_common_prefixes(bucket, prefix):
    """
    Helper function to get common "subfolders" of folders
    in S3

    Args:
        bucket (str): bucket name
        prefix (str): prefix for which to list common prefixes

    Returns:

    """
    if not prefix.endswith('/'):
        prefix += "/"
    client = boto3.client('s3')
    paginator = client.get_paginator('list_objects')
    result = paginator.paginate(Bucket=bucket, Delimiter='/', Prefix=prefix)
    return [common_prefix['Prefix'].split('/')[-2]
            for common_prefix in result.search("CommonPrefixes")
            if common_prefix]


def main():
    """
    Main function for running this as a script

    Returns:
        None
    """
    args = docopt(__doc__)
    campaign = args['--campaign']
    catch_errors = args['--catch_errors']
    worker = Worker(campaign, catch_errors=catch_errors)
    num_loops = args['--loops']
    if num_loops is not None:
        num_loops = int(num_loops)
    else:
        num_loops = np.inf
    if args['COMMAND'] == "start":
        print("Starting {} worker with {} loops".format(
            campaign, num_loops))
        worker.remove_stop_file()
        worker.start(num_loops=num_loops)
    elif args['COMMAND'] == "stop":
        worker.write_stop_file()
    else:
        raise ValueError("Invalid command {}.  Worker command must be 'start' or 'stop'")


if __name__ == "__main__":
    main()



"""This module defines agents and various functions
that use megnet for candidate selection or candidate filtration
"""
import numpy as np
import tensorflow as tf
from multiprocessing import cpu_count

from m3gnet.models import M3GNet, Potential, Relaxer
from m3gnet.trainers import PotentialTrainer

from camd.agent.stability import StabilityAgent
from camd.agent.base import HypothesisAgent


def reverse_calcs(retrain_data, data_screening=None):
    """
    Prepare the data for training
    Args:
        retrain_data (pd.DataFrame):
        data_screening (dict): retrain data select criteria
                            example: {"task": "relax1",
                                    "ionic_step": "last_n"}
    Returns: lists
        X_struct, Xe, Xf, Xs
    """
    X_struct = []
    Xe = []
    Xf = []
    Xs = []
    for d in retrain_data:
        for task in d:
            for step, ionic_steps in enumerate(task["output"]["ionic_steps"]):
                # TODO: implement data selection criteria
                X_struct.append(ionic_steps["structure"])
                Xe.append(
                    ionic_steps["e_fr_energy"] / ionic_steps["structure"].num_sites
                )
                Xf.append(ionic_steps["forces"])
                Xs.append(ionic_steps["stress"])
    return X_struct, Xe, Xf, Xs


class M3GNetStabilityAgent(StabilityAgent):
    """Stability agent that uses megnet for estimation of energy"""

    def __init__(
        self,
        m3gnet=None,
        learning_rate=1e-3,
        candidate_data=None,
        seed_data=None,
        n_query=1,
        hull_distance=0.0,
        parallel=cpu_count(),
    ):
        """


        Args:
            m3gnet (M3gnet): m3gnet object, graph CNN from m3gnet package
                which determines energy, forces, stresses, and can do relaxation
                etc.
            learning_rate (float): learning rate for m3gnet fitting
            candidate_data (pd.DataFrame): candidate data to predict on
            seed_data (pd.DataFrame): seed data to fit on
            n_query (int): number of candidates to select
            hull_distance (float): hull distance to consider
            parallel (int): number of parallel processes to use for
                phase determination
        """
        super(M3GNetStabilityAgent, self).__init__(
            candidate_data=candidate_data,
            seed_data=seed_data,
            n_query=n_query,
            hull_distance=hull_distance,
            parallel=parallel,
        )
        if not m3gnet:
            self.m3gnet = M3GNet(is_intensive=False)
        else:
            self.m3gnet = m3gnet
        self.potential = Potential(model=self.m3gnet)
        self.trainer = PotentialTrainer(
            potential=self.potential, optimizer=tf.keras.optimizers.Adam(learning_rate)
        )

    def get_hypotheses(
        self, candidate_data, seed_data=None, retrain_committee=False, retrain_epochs=1
    ):
        """
        Get hypotheses method for downselecting candidates

        Args:
            candidate_data (pd.DataFrame): candidate data to predict on and select from
            seed_data (pd.DataFrame): seed data to fit on
            retrain_committee (bool): whether to retrain m3gnet on each run
            retrain_epochs (int): number of epochs for retraining

        Returns:
            (pd.DataFrame): selected candidates for experiment

        """
        if "target" in candidate_data.columns:
            self.candidate_data = candidate_data.drop(columns=["target"], axis=1)
        else:
            self.candidate_data = candidate_data
        # TODO: handle this here for now, should move later, as in stability agents
        self.seed_data = seed_data.dropna(subset=['delta_e'])
        X_seed = seed_data.drop(columns=["target"], axis=1, errors="ignore")
        # y_seed = seed_data["target"]
        if retrain_committee and "calcs_reversed" in X_seed.columns:
            self.train(X_seed, retrain_epochs)
        relaxer = Relaxer(potential=self.trainer.potential)
        t_pred = []
        for s in candidate_data["structure"]:
            t = relaxer.relax(s)["trajectory"]
            e = t.energies[-1].flatten()[0]
            t_pred.append(e)
        # TODO: get real formation energy instead of energy per atom
        self.update_candidate_stabilities(t_pred, sort=True, floor=-6.0)
        # Find the most stable ones up to n_query within hull_distance
        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
        within_hull = self.candidate_data[stability_filter]
        return within_hull.head(self.n_query)

    def train(self, train_data, epochs=10, kwargs={}):
        """
        Train the potential
        Args:
            train_data (pd.DataFrame): data for re-train,
                                        same format as seed data
                                        should contain "calcs_reversed" column
            epochs (int): number of max epoch to train
            kwargs (dict): parameters for trainer.train

        Returns:
            None
        """
        X_struct, Xe, Xf, Xs = reverse_calcs(train_data["calcs_reversed"])
        self.trainer.train(
            graphs_or_structures=X_struct,
            energies=Xe,
            forces=Xf,
            stresses=Xs,
            epochs=epochs,
            **kwargs
        )


class M3GNetHypothesisAgent(HypothesisAgent):
    """Generic agent for AL using m3gnet"""

    def __init__(self, m3gnet=None, candidate_data=None, seed_data=None, n_query=None):

        """
        Args:
            m3gnet (M3gnet): m3gnet object, graph CNN from m3gnet package
                which determines energy, forces, stresses, and can do relaxation
                etc.
            candidate_data (pd.DataFrame): candidate data to predict on
            seed_data (pd.DataFrame): seed data to fit on
            n_query (int): number of candidates to select
        """
        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.n_query = n_query if n_query else 1
        # self.cv_score = np.inf
        if not m3gnet:
            self.m3gnet = M3GNet(is_intensive=False)
        else:
            self.m3gnet = m3gnet
        self.potential = Potential(model=self.m3gnet)
        self.trainer = PotentialTrainer(
            potential=self.potential, optimizer=tf.keras.optimizers.Adam(1e-3)
        )
        super(M3GNetHypothesisAgent).__init__()

    def get_hypotheses(self, candidate_data, seed_data=None, retrain=False):

        """
        Get hypotheses method for downselecting candidates

        Args:
            candidate_data (pd.DataFrame): candidate data to predict on and select from
            seed_data (pd.DataFrame): seed data to fit on
            retrain (bool): whether to retrain m3gnet on each run

        Returns:
            (pd.DataFrame): selected candidates for experiment

        """
        if "target" in candidate_data.columns:
            self.candidate_data = candidate_data.drop(columns=["target"], axis=1)
        else:
            self.candidate_data = candidate_data
        self.seed_data = seed_data
        X_seed = seed_data.drop(columns=["target"], axis=1, errors="ignore")
        # y_seed = seed_data["target"]
        if retrain and "calcs_reversed" in X_seed.columns:
            self.train(X_seed)
        relaxer = Relaxer(potential=self.trainer.potential)
        t_pred = []
        for s in self.candidate_data["structure"]:
            t = relaxer.relax(s)["trajectory"]
            e = t.energies[-1].flatten()[0]
            t_pred.append(e)
        selected = np.argsort(-1.0 * np.array(t_pred))[: self.n_query]
        return candidate_data.iloc[selected]

    def train(self, train_data, epochs=10, kwargs=None):
        """
        Train the potential

        Args
            train_data (pd.DataFrame): data for re-train,
                                        same format as seed data
                                        should contain "calcs_reversed" column
            epochs (int): number of max epoch to train
            kwargs (dict): parameters for trainer.train
        Returns:
            None
        """
        if kwargs is None:
            kwargs = {}
        X_struct, Xe, Xf, Xs = reverse_calcs(train_data["calcs_reversed"])
        self.trainer.train(
            graphs_or_structures=X_struct,
            energies=Xe,
            forces=Xf,
            stresses=Xs,
            epochs=epochs,
            **kwargs
        )



# Copyright Toyota Research Institute 2021
"""
This module implements agents and helper functions designed
support multi-fidelity discovery campaigns.

See the following reference:

Palizhati A, Aykol M, Suram S, Hummelshøj JS, Montoya JH. Multi-fidelity Sequential Learning for
Accelerated Materials Discovery. ChemRxiv. Cambridge: Cambridge Open Engage; 2021; This content
is a preprint and has not been peer-reviewed.
"""


import GPy
import numpy as np
import pandas as pd

from sklearn.preprocessing import StandardScaler
from camd.agent.base import HypothesisAgent


def get_features_from_df(df, features):
    """
    Helper function to get feature columns of a dataframe.

    Args:
        df (pd.DataFrame)       A pd.DataFrame where the features are extracted.
        features (list of str)  Name of the features columns in df.

    Returns:
            feature_df      A pd.DataFrame that only contains the features used in ML.
    """
    feature_df = df[features]
    return feature_df


def process_data(
    candidate_data, seed_data, features, label, y_reshape=False, preprocessor=None
):
    """
    Helper function that process data for ML model and returns
    np.ndarray of training features, training labels,
    test features and test labels.
    """
    X_train = get_features_from_df(seed_data, features).values
    y_train = np.array(seed_data[label])
    X_test = get_features_from_df(candidate_data, features).values
    y_test = np.array(candidate_data[label])

    if y_reshape:
        y_train = y_train.reshape(-1, 1)
        y_test = y_test.reshape(-1, 1)

    if preprocessor:
        X_train = preprocessor.fit_transform(X_train)
        X_test = preprocessor.transform(X_test)
    return X_train, y_train, X_test, y_test


class EpsilonGreedyMultiAgent(HypothesisAgent):
    """
    A multi-fidelity agent that allocates the desired budget for high fidelity versus
    low fidelity candidate data, and acquires candidates in each fidelity via exploitation
    """

    def __init__(
        self,
        candidate_data=None,
        seed_data=None,
        features=None,
        fidelities=("theory_data", "expt_data"),
        target_prop=None,
        target_prop_val=None,
        preprocessor=StandardScaler(),
        model=None,
        ranking_method="minimize",
        total_budget=None,
        highFi_query_frac=None,
        similarity_thres=300.0,
        lowFi_per_highFi=1,
    ):
        """
        Args:
            candidate_data (df)      Candidate search space for the campaign.
            seed_data (df)           Seed (training) data for the campaign.
            features (tuple of str)  Name of the feature columns used in machine learning model training.
            fidelities (tuple)       Fidelity levels of the datasets. The strings in the tuple should be arranged
                                     from low to high fidelity.
            target_prop (str)        The property machine learning model is learning, given feature space.
            target_prop_val (float)  The ideal value of the target property.
            preprocessor             A preprocessor that preprocess the features. It can be None, a single
                                     processor, or a pipeline. The default is sklearn StandardScaler.
            model                    A sklearn supervised machine learning regressor.
            ranking_method (str)     Ranking method of candidates based on ML predictions. Select one of the
                                     following: "minimize", "ascending", or "descending". "minimize" will rank
                                     candidates with ML candidates closest to the target property value. "ascending"
                                     or "descening" will rank candidates with ascending/descening ML predictions, best
                                     to use when there is no target propety value (i.e. smaller/larger the better).
            total_budget (int)       The number of the hypotheses at a given iteration of the campaign.
            highFi_query_frac        The fraction of the total budget used for high fidelity hypotheses queries.
                                     The value should be <0 and <=1.
            similarity_thres(float)  The threshold value for l2 norm similarity between a candidate composition and
                                     compositions in the seed data. User will need to run some calculations
                                     to determine the best threshold value.
            lowFi_per_highFi (int)   The number of low fidelity candidate selected to support each
                                     high fidelity candidates that predicted to be good, but the agent
                                     does not want to generate that experimental hypotheses yet.
        """
        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.features = features
        self.fidelities = fidelities
        self.target_prop = target_prop
        self.target_prop_val = target_prop_val
        self.preprocessor = preprocessor
        self.model = model
        self.ranking_method = ranking_method
        self.total_budget = total_budget
        self.highFi_query_frac = highFi_query_frac
        self.similarity_thres = similarity_thres
        self.lowFi_per_highFi = lowFi_per_highFi
        super(EpsilonGreedyMultiAgent).__init__()

    def _calculate_similarity(self, comp, seed_comps):
        """
        Helper function that calculates similarity between
        a composition and the seed data compositions. The similarity
        is reprsented by l2_norm.

        Args:
            comp(pd.core.series):    A specific composition represented by Magpie.
            seed_comps (df):         Compostions in seed represented by Magpie.
        """
        l2_norm = np.linalg.norm(comp.values - seed_comps.values, axis=1)
        return l2_norm

    def _query_hypotheses(self, candidate_data, seed_data):
        """
        Query hypotheses given candidate and seed data via exploitation.
        """
        # separate the candidate space into high and low fidelity candidates
        high_fidelity_candidates = candidate_data.loc[
            candidate_data[self.fidelities[1]] == 1
        ]
        low_fidelity_candidates = candidate_data.loc[
            candidate_data[self.fidelities[0]] == 1
        ]

        # edge cases: end campaign if there are no high fidelity candidate
        # use the entire query budget if there are only high fidelity candidate
        if len(high_fidelity_candidates) == 0:
            return None

        elif (len(high_fidelity_candidates) != 0) & (len(low_fidelity_candidates) == 0):
            selected_hypotheses = high_fidelity_candidates.head(self.total_budget)

        else:
            selected_hypotheses = pd.DataFrame(columns=candidate_data.columns)

            # Query high fidelity candidates first
            highFi_budget = int(self.total_budget * self.highFi_query_frac)
            seed_data_fea = get_features_from_df(seed_data, self.features)
            highFi_cands_fea = get_features_from_df(
                high_fidelity_candidates, self.features
            )

            for idx, cand_fea in highFi_cands_fea.iterrows():
                if len(selected_hypotheses) < highFi_budget:
                    normdiff = self._calculate_similarity(cand_fea, seed_data_fea)
                    if len(normdiff[(normdiff <= self.similarity_thres)]) >= 1:
                        selected_hypotheses = selected_hypotheses.append(
                            high_fidelity_candidates.loc[idx]
                        )

            # query low fidelity candidate for remaining budget
            remained_highFi_cands_fea = highFi_cands_fea.drop(selected_hypotheses.index)
            lowFi_candidates_copy = low_fidelity_candidates.copy()
            for idx, cand_fea in remained_highFi_cands_fea.iterrows():
                if (len(selected_hypotheses) < self.total_budget) & (
                    len(lowFi_candidates_copy) != 0
                ):
                    lowFi_cands_fea = get_features_from_df(
                        lowFi_candidates_copy, self.features
                    )
                    lowFi_candidates_copy["normdiff"] = self._calculate_similarity(
                        cand_fea, lowFi_cands_fea
                    )
                    lowFi_candidates_copy = lowFi_candidates_copy.sort_values(
                        "normdiff"
                    )
                    selected_hypotheses = selected_hypotheses.append(
                        lowFi_candidates_copy.head(self.lowFi_per_highFi)
                    )
                    lowFi_candidates_copy = lowFi_candidates_copy.drop(
                        lowFi_candidates_copy.head(self.lowFi_per_highFi).index
                    )
        return selected_hypotheses

    def get_hypotheses(self, candidate_data, seed_data):
        """
        Gets hypotheses using agent.

        Args:
            candidate_data (pd.DataFrame): dataframe of candidates
            seed_data (pd.DataFrame): dataframe of known data

        Returns:
            (pd.DataFrame): dataframe of selected candidates

        """
        features_columns = self.features + list(self.fidelities)
        X_train, y_train, X_test, y_test = process_data(
            candidate_data,
            seed_data,
            features_columns,
            self.target_prop,
            preprocessor=self.preprocessor,
        )
        self.model.fit(X_train, y_train)
        y_pred = self.model.predict(X_test)

        # Make a copy of the candidate data so the original one
        # does not get modified during hypotheses generation
        candidate_data_copy = candidate_data.copy()
        candidate_data_copy["distance_to_ideal"] = np.abs(self.target_prop_val - y_pred)

        if self.ranking_method == "minimize":
            candidate_data_copy = candidate_data_copy.sort_values(
                by=["distance_to_ideal"]
            )
        elif self.ranking_method == "ascending":
            candidate_data_copy = candidate_data_copy.sort_values(
                by=y_pred, ascending=True
            )
        elif self.ranking_method == "descending":
            candidate_data_copy = candidate_data_copy.sort_values(
                by=y_pred, ascending=False
            )

        hypotheses = self._query_hypotheses(candidate_data_copy, seed_data)
        return hypotheses


class GPMultiAgent(HypothesisAgent):
    """
    A Gaussian process lower confidence bound derived multi-fidelity agent.
    This agent operates under a total acquisition budget. It acquires
    candidates factoring in GPR predicted uncertainties in the LCB setting
    and hallucination of information gain from DFT acquisitions analogous
    to work of Desautels et al. in batch mode LCB.  The agent aims for
    prioritizing exploitation primarily with high-fidelity experimental measurements,
    offloading exploratory (higher risk) acquisitions first to lower-fidelity computations.
    """

    def __init__(
        self,
        candidate_data=None,
        seed_data=None,
        chemsys_col="reduced_formula",
        features=None,
        fidelities=("theory_data", "expt_data"),
        target_prop=None,
        target_prop_val=1.8,
        total_budget=None,
        preprocessor=StandardScaler(),
        gp_max_iter=200,
        alpha=1.0,
        rank_thres=10,
        unc_percentile=5,
    ):
        """
        Args:
            candidate_data (df)      Candidate search space for the campaign.
            seed_data (df)           Seed (training) data for the campaign.
            chemsys_col (str)        The name of the composition column.
            features (tuple of str)  Column name of the features used in machine learning.
            fidelities (tuple)       Fidelity levels of the datasets. The strings in the tuple should be arranged
                                     from low to high fidelity. The value of fidelity features should
                                     be one-hot encoded.
            target_prop (str)        The property machine learning model is predicting, given feature space.
            target_prop_val (float)  The ideal value of the target property.
            total_budget (int)       The budget for the hypotheses queried.
            gp_max_iter (int)        Number of maximum iterations for GP optimization.
            preprocessor             A preprocessor that preprocess the features. It can be None, a single
                                     processor, or a pipeline. The default is StandardScaler().
            alpha (float)            The mixing parameter for uncertainties. It controls the
                                     trade-off between exploration and exploitation. Defaults to 1.0.
            rank_thres (int)         A threshold help to decide if lower fidelity data is worth acquiring.
            unc_percentile (int)     A number between 0 and 100, and used to calculate an uncertainty threshold
                                     at that percentile value. The threshold is used to decide if the
                                     agent is quering theory or experimental hypotheses.
        """
        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.chemsys_col = chemsys_col
        self.features = features
        self.fidelities = fidelities
        self.target_prop = target_prop
        self.target_prop_val = target_prop_val
        self.total_budget = total_budget
        self.preprocessor = preprocessor
        self.gp_max_iter = gp_max_iter
        self.alpha = alpha
        self.rank_thres = rank_thres
        self.unc_percentile = unc_percentile
        super(GPMultiAgent).__init__()

    def _halluciate_lower_fidelity(self, seed_data, candidate_data, low_fidelity_data):
        # make copy of seed and candidate data, so we don't mess with the original one
        new_candidate_data = candidate_data.copy()
        low_fidelity = low_fidelity_data.copy()

        low_fidelity[self.target_prop] = low_fidelity_data["y_pred"]
        low_fidelity = low_fidelity.drop(columns=["y_pred"])
        new_seed_data = pd.concat([seed_data, low_fidelity])
        new_candidate_data = new_candidate_data.drop(low_fidelity.index)
        pred_candidate_data = self._train_and_predict(new_candidate_data, new_seed_data)
        return pred_candidate_data

    def _get_rank_after_hallucination(
        self, seed_data, candidate_data, orig_idx, low_fidelity
    ):
        halluciated_candidates = self._halluciate_lower_fidelity(
            seed_data, candidate_data, low_fidelity
        )
        halluciated_candidates = halluciated_candidates.loc[
            halluciated_candidates[self.fidelities[1]] == 1
        ]
        halluciated_candidates = halluciated_candidates.sort_values("pred_lcb")
        rank_after_hallucination = halluciated_candidates.index.get_loc(orig_idx)
        return rank_after_hallucination

    def _train_and_predict(self, candidate_data, seed_data):
        features_columns = self.features + list(self.fidelities)
        X_train, y_train, X_test, y_test = process_data(
            candidate_data,
            seed_data,
            features_columns,
            self.target_prop,
            y_reshape=True,
            preprocessor=self.preprocessor,
        )
        gp = GPy.models.GPRegression(X_train, y_train)
        gp.optimize("bfgs", max_iters=self.gp_max_iter)
        y_pred, var = gp.predict(X_test)

        # Make a copy of the candidate data so the original one
        # does not get modified during hypotheses generation
        candidate_data_copy = candidate_data.copy()
        dist_to_ideal = np.abs(self.target_prop_val - y_pred)
        pred_lcb = dist_to_ideal - self.alpha * var**0.5
        candidate_data_copy["pred_lcb"] = pred_lcb
        candidate_data_copy["pred_unc"] = var**0.5
        candidate_data_copy["y_pred"] = y_pred
        return candidate_data_copy

    def _query_hypotheses(self, candidate_data, seed_data):
        high_fidelity_candidates = candidate_data.loc[
            candidate_data[self.fidelities[1]] == 1
        ]
        high_fidelity_candidates = high_fidelity_candidates.sort_values("pred_lcb")

        # edge case: top the campaign if there are no high fidelity candidates
        if len(high_fidelity_candidates) == 0:
            return None

        else:
            selected_hypotheses = pd.DataFrame(columns=candidate_data.columns)
            unc_thres = np.percentile(
                np.array(high_fidelity_candidates.pred_unc), self.unc_percentile
            )

            # query hypothesis
            for idx, candidate in high_fidelity_candidates.iterrows():
                if len(selected_hypotheses) < self.total_budget:
                    chemsys = candidate[self.chemsys_col]
                    low_fidelity = candidate_data.loc[
                        (candidate_data[self.chemsys_col] == chemsys)
                        & (candidate_data[self.fidelities[0]] == 1)
                    ]

                    # exploit if uncertainty is low or the low fidelity data is not available
                    if (candidate.pred_unc <= unc_thres) or (len(low_fidelity) == 0):
                        selected_hypotheses = selected_hypotheses.append(candidate)
                    # explore
                    else:
                        orig_rank = high_fidelity_candidates.index.get_loc(idx)
                        new_rank = self._get_rank_after_hallucination(
                            seed_data, candidate_data, idx, low_fidelity
                        )

                        delta_rank = new_rank - orig_rank
                        if delta_rank <= self.rank_thres:
                            selected_hypotheses = selected_hypotheses.append(candidate)
                        else:
                            selected_hypotheses = selected_hypotheses.append(
                                low_fidelity
                            )
        return selected_hypotheses

    def get_hypotheses(self, candidate_data, seed_data):
        """
        Selects candidate data for experiment using agent methods

        Args:
            candidate_data (pd.DataFrame): candidate data
            seed_data (pd.DataFrame): known data on which to base the selection
                of candidates

        Returns:
            (pd.DataFrame): selected candidates e.g. for experiment

        """
        candidate_data = self._train_and_predict(candidate_data, seed_data)
        hypotheses = self._query_hypotheses(candidate_data, seed_data)
        return hypotheses



# Copyright Toyota Research Institute 2019
"""
Contains basic logic and abstraction for experiments and
generic experiments like the ATF Sampler.
"""

import abc

import pandas as pd
from monty.json import MSONable


# TODO: rethink naming here, should it be ExperimentHandler?
# TODO: rethink history storage
class Experiment(abc.ABC, MSONable):
    """
    An abstract class for brokering experiments, this is a
    temporary placeholder for the way we do experiments

    Eventually, this step should execute asynchronously,
    we might just want to use FireWorks for this
    """

    def __init__(self, current_data=None, job_status=None, history=None):
        """
        Initializes an experiment.

        Args:
            current_data (pandas.DataFrame): dataframe corresponding to
                currently submitted experiments
            job_status (str): status of the experiment
            history (pandas.DataFrame): history of past experiments
        """
        self.current_data = current_data
        self.job_status = job_status
        self._history = history or []

    def update_current_data(self, data):
        """
        Updates current data with dataframe,
        stores old data in history

        Args:
            data (DataFrame):

        Returns:
            None

        """
        if self.current_data is not None:
            current_results = self.get_results()
            self._history.append((self.current_data, current_results))

        self.current_data = data

    @abc.abstractmethod
    def get_results(self):
        """
        Returns:
            dict: a dictionary of results

        """

    @abc.abstractmethod
    def monitor(self):
        """
        Keeps track of jobs given the poll_time and timeout
        """

    @abc.abstractmethod
    def submit(self, data):
        """
        Args:
            data (DataFrame): dataframe containing all necessary
                data to conduct the experiment(s).  May be one
                row, may be multiple rows

        Returns:
            None
        """

    @property
    def agg_history(self):
        """
        Aggregated history, i.e. in two single dataframes
        corresponding to "current data" attributes and
        results

        Returns:
            (DataFrame): history of current data
            (DataFrame): history of results

        """
        cd_list, cr_list = zip(*self._history)
        return pd.concat(cd_list), pd.concat(cr_list)


class ATFSampler(Experiment):
    """
    A simple after the fact sampler that just samples
    a dataframe according to index_values
    """

    def __init__(self, dataframe, current_data=None, job_status=None):
        """
        Initializes an ATFSampler Experiment

        Args:
            dataframe (pandas.DataFrame): dataframe of a-priori results
                from which to pull
            current_data (pandas.DataFrame): data of currently submitted
                "experiments"
            job_status (str): status of job

        """
        self.dataframe = dataframe
        super(ATFSampler, self).__init__(
            current_data=current_data, job_status=job_status
        )

    def get_results(self):
        """
        Simply samples the dataframe associated with the ATFSampler
        object according to the last submitted data

        Returns:
            (DataFrame): DataFrame of results

        """
        indices = self.current_data.index
        return self.dataframe.loc[indices]

    def submit(self, data):
        """
        This does nothing other than update current_data,
        since the "experiments" are already done
        """
        self.update_current_data(data)
        self.job_status = "COMPLETED"
        return None

    def monitor(self):
        """
        ATF Sampler returns results synchronously,
        and thus needs no monitor, so this simply
        returns True

        Returns:
            True

        """
        return True



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.

import unittest
import pandas as pd
import os
import boto3
import json

from monty.tempfile import ScratchDir
from camd import CAMD_TEST_FILES, CAMD_S3_BUCKET
from camd.campaigns.base import Campaign
from camd.agent.stability import AgentStabilityML5
from camd.analysis import StabilityAnalyzer
from camd.experiment.base import ATFSampler


def teardown_s3():
    """Tear down test files in s3"""
    s3 = boto3.resource('s3')
    bucket = s3.Bucket(CAMD_S3_BUCKET)
    bucket.objects.filter(Prefix="{}".format("test")).delete()


class CampaignTest(unittest.TestCase):
    def tearDown(self):
        teardown_s3()

    @unittest.skipUnless(CAMD_S3_BUCKET, "CAMD S3 Bucket not set")
    def test_sync(self):
        with ScratchDir('.'):
            df = pd.read_csv(os.path.join(CAMD_TEST_FILES, 'test_df.csv'))

            # Construct and start campaign
            new_campaign = Campaign(df, AgentStabilityML5(), ATFSampler(df),
                                    StabilityAnalyzer(), create_seed=10,
                                    s3_prefix="test")
            new_campaign.auto_loop(n_iterations=3, save_iterations=True, initialize=True)
        # Test iteration read
        s3 = boto3.resource('s3')
        obj = s3.Object(CAMD_S3_BUCKET, "test/iteration.json")
        loaded = json.loads(obj.get()['Body'].read())
        self.assertEqual(loaded, 2)

        # Test save directories
        for iteration in [-1, 0, 1, 2]:
            obj = s3.Object(CAMD_S3_BUCKET, f"test/{iteration}/iteration.json")
            loaded = json.loads(obj.get()['Body'].read())
            self.assertEqual(loaded, iteration)


if __name__ == '__main__':
    unittest.main()



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.

import unittest
import pandas as pd
from monty.tempfile import ScratchDir
from camd.utils.data import (
    load_dataframe,
    partition_intercomp,
    get_oqmd_data_by_chemsys,
)
from camd.agent.base import RandomAgent
from camd.experiment.agent_simulation import LocalAgentSimulation
from camd.analysis import StabilityAnalyzer


class CampaignSimulationTest(unittest.TestCase):
    def test_run(self):
        with ScratchDir("."):
            dataframe = get_oqmd_data_by_chemsys("Fe-O")
            cand, seed = partition_intercomp(dataframe, n_elements=1)
            agents_df = pd.DataFrame({"agent": [RandomAgent()]})
            simulation = LocalAgentSimulation(
                cand, iterations=5, seed_data=seed, analyzer=StabilityAnalyzer()
            )
            simulation.submit(agents_df)
            simulation.monitor()
            results = simulation.get_results()
            self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.
"""
Module and script for running the CAMD campaign runner

Usage:
    camd_runner CAMPAIGN_TYPE CAMPAIGN_NAME [options]

Options:
    --scratch            run in scratch directry
    -h --help            Show this screen
    --version            Show version

CAMPAIGN_TYPE should be a "type" for a campaign,
i.e. meta-agent or proto-dft-2, CAMPAIGN_ARG,
should be an input to the campaign e.g. Si-Ag
for chemsys or a reserved name for meta-agents
"""

import os
import tempfile
import shutil
from docopt import docopt
from camd.campaigns.meta_agent import MetaAgentCampaign
from camd.campaigns.structure_discovery import ProtoDFTCampaign


def main():
    """Utility function to run a reserved name campaign"""
    args = docopt(__doc__)
    campaign_type = args['CAMPAIGN_TYPE']
    campaign_name = args['CAMPAIGN_NAME']
    cwd = os.getcwd()
    if args['--scratch']:
        dirpath = tempfile.mkdtemp()
        os.chdir(dirpath)

    # Switch for different campaign types
    if campaign_type.startswith("proto-dft"):
        prefix = "{}/runs".format(campaign_type)
        campaign = ProtoDFTCampaign.from_chemsys(campaign_name, prefix=prefix)
    elif campaign_type == "meta_agent":
        campaign = MetaAgentCampaign.from_reserved_name(campaign_name)
    else:
        raise ValueError("{} is not a supported campaign type".format(
            campaign_type))

    campaign.autorun()

    # Cleanup
    if args['--scratch']:
        os.chdir(cwd)
        shutil.rmtree(dirpath)


if __name__ == "__main__":
    main()



import unittest
import datetime

from camd.utils import get_new_version


class UtilsTest(unittest.TestCase):
    def test_get_new_version(self):
        today = datetime.datetime.today().strftime("%Y.%-m.%-d")
        # New version is new date
        new_version = get_new_version(current_ver="2020.1.1")
        self.assertEqual(new_version, today)

        # New version is first "post version"
        new_version = get_new_version(current_ver=today)
        self.assertEqual(new_version, "{}-post0".format(today))

        # New version is incrementing "post version"
        new_version = get_new_version(current_ver="{}-post3".format(today))
        self.assertEqual(new_version, "{}-post4".format(today))

        # New version increments multiple digits for "post version"
        new_version = get_new_version(current_ver="{}-post10".format(today))
        self.assertEqual(new_version, "{}-post11".format(today))


if __name__ == '__main__':
    unittest.main()



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.

import unittest
import os
import uuid

from pymatgen.util.testing import PymatgenTest
from pymatgen.ext.matproj import MPRester
from camd.experiment.dft import OqmdDFTonMC1
import pandas as pd


CAMD_DFT_TESTS = os.environ.get("CAMD_DFT_TESTS", False)
SKIP_MSG = "Long tests disabled, set CAMD_DFT_TESTS to run long tests"


# This test is still inconsistent because of issues with
# batch AWS jobs and database communications
class Mc1Test(unittest.TestCase):
    @unittest.skipUnless(CAMD_DFT_TESTS, SKIP_MSG)
    def test_get(self):
        test_uuid = uuid.uuid4()
        good_silicon = PymatgenTest.get_structure("Si")
        bad_silicon = good_silicon.copy()

        # Add another site at the same position
        bad_silicon.append("Si", [0.1, 0.1, 0.15])
        bad_silicon.append("Si", [0.1, 0.333, 0.15])
        self.assertEqual(len(bad_silicon), 4)
        data = pd.DataFrame({"structure": {"good": good_silicon, "bad": bad_silicon}})

        experiment = OqmdDFTonMC1(
            poll_time=30, timeout=150, prefix_append=str(test_uuid)
        )
        experiment.submit(data)
        status = experiment.monitor()
        results = experiment.get_results()

        self.assertAlmostEqual(results.loc["good", "delta_e"], 0, 5)
        self.assertIsNone(results.loc["bad", "result"])

        # Test cleanup functionality
        experiment.poll_time = 1
        experiment.timeout = 1
        old_paths = experiment.current_data["path"]
        data = pd.DataFrame(
            {"structure": {"newgood": good_silicon, "newbad": bad_silicon}}
        )
        experiment.submit(data)
        experiment.monitor()
        for path in old_paths:
            self.assertFalse(os.path.exists(path.replace("model", "simulation")))

    @unittest.skipUnless(CAMD_DFT_TESTS, SKIP_MSG)
    def test_caching(self):
        good_silicon = PymatgenTest.get_structure("Si")
        bad_silicon = good_silicon.copy()

        # Add another site at the same position
        bad_silicon.append("Si", [0.1, 0.1, 0.15])
        bad_silicon.append("Si", [0.1, 0.333, 0.15])
        self.assertEqual(len(bad_silicon), 4)
        data = pd.DataFrame(
            {
                "structure": {
                    "good": good_silicon,
                    "bad": bad_silicon,
                }
            }
        )
        experiment = OqmdDFTonMC1(
            poll_time=30, timeout=150, prefix_append="camd-cache-test"
        )
        # This generates the data if needed in the future
        # experiment.submit(data)
        # status = experiment.monitor()
        # results = experiment.get_results()

        experiment.update_current_data(data)
        results = experiment.fetch_cached(data)

        self.assertAlmostEqual(results.loc["good", "delta_e"], 0, 5)
        self.assertIsNone(results.loc["bad", "result"])
        self.assertEqual(results.loc["bad", "status"], "FAILED")

    @unittest.skipUnless(CAMD_DFT_TESTS, SKIP_MSG)
    def test_structure_suite(self):
        mp_ids = ["mp-702", "mp-1953", "mp-1132", "mp-8409", "mp-872"]
        with MPRester() as mpr:
            structure_dict = {
                mp_id: mpr.get_structure_by_material_id(mp_id) for mp_id in mp_ids
            }
        data = pd.DataFrame({"structure": structure_dict})

        experiment = OqmdDFTonMC1(poll_time=25)
        experiment.submit(data)
        status = experiment.monitor()
        results = experiment.get_results()
        self.assertTrue(all(results["status"] == "SUCCEEDED"))


if __name__ == "__main__":
    unittest.main()



#  Copyright (c) 2021 Toyota Research Institute.  All rights reserved.
"""
Module that defines logging utilities, e.g. for stdout
and AWS Kinesis
"""
import boto3
import logging


class KinesisStreamHandler(logging.StreamHandler):
    def __init__(self, stream_name, partition_key):
        # By default, logging.StreamHandler uses sys.stderr
        # if stream parameter is not specified
        logging.StreamHandler.__init__(self)
        self.__kinesis = None
        self.__stream_buffer = []
        try:
            self.__kinesis = boto3.client('kinesis')
        except Exception:
            print('Kinesis client initialization failed.')
        self.__stream_name = stream_name
        self.__partition_key = partition_key

    def emit(self, record):
        try:
            msg = self.format(record)
            if self.__kinesis:
                self.__stream_buffer.append(
                    msg.encode(encoding="UTF-8", errors="strict")
                )
            else:
                stream = self.stream
                stream.write(msg)
                stream.write(self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)

    def flush(self):
        self.acquire()
        try:
            if self.__kinesis and self.__stream_buffer:
                import pdb; pdb.set_trace()
                for record in self.__stream_buffer:
                    self.__kinesis.put_record(
                        StreamName=self.__stream_name,
                        PartitionKey=self.__partition_key,
                        Data=record
                    )
                self.__stream_buffer.clear()
        except Exception as e:
            print("An error occurred during flush operation.")
            print(f"Exception: {e}")
            print(f"Stream buffer: {self.__stream_buffer}")
        finally:
            if self.stream and hasattr(self.stream, "flush"):
                self.stream.flush()
            self.release()


class CloudWatchHandler(logging.StreamHandler):
    def __init__(self, stream_name, partition_key):
        # By default, logging.StreamHandler uses sys.stderr
        # if stream parameter is not specified
        logging.StreamHandler.__init__(self)
        self.__kinesis = None
        self.__stream_buffer = []
        try:
            self.__kinesis = boto3.client('kinesis')
        except Exception:
            print('Kinesis client initialization failed.')
        self.__stream_name = stream_name
        self.__partition_key = partition_key

    def emit(self, record):
        try:
            msg = self.format(record)
            if self.__kinesis:
                self.__stream_buffer.append(
                    msg.encode(encoding="UTF-8", errors="strict")
                )
            else:
                stream = self.stream
                stream.write(msg)
                stream.write(self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)

    def flush(self):
        self.acquire()
        try:
            if self.__kinesis and self.__stream_buffer:
                import pdb; pdb.set_trace()
                for record in self.__stream_buffer:
                    self.__kinesis.put_record(
                        StreamName=self.__stream_name,
                        PartitionKey=self.__partition_key,
                        Data=record
                    )
                self.__stream_buffer.clear()
        except Exception as e:
            print("An error occurred during flush operation.")
            print(f"Exception: {e}")
            print(f"Stream buffer: {self.__stream_buffer}")
        finally:
            if self.stream and hasattr(self.stream, "flush"):
                self.stream.flush()
            self.release()



import unittest
import os
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from m3gnet.models import M3GNet

from camd import CAMD_TEST_FILES
from camd.agent.m3gnet import M3GNetHypothesisAgent, M3GNetStabilityAgent, reverse_calcs


class HypothesisAgentsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        test_data = pd.read_pickle(os.path.join(CAMD_TEST_FILES, "test_m3gnet.pickle"))
        cls.test_data = test_data
        cls.seed_data, cls.candidate_data = train_test_split(
            test_data, train_size=0.2, random_state=42
        )
        model = M3GNet(is_intensive=False)
        cls.agent = M3GNetHypothesisAgent(m3gnet=model)

    def test_reverse_calcs(self):
        X_struct, Xe, Xf, Xs = reverse_calcs(self.test_data["calcs_reversed"].iloc[:2])
        self.assertEqual(len(X_struct), 90)
        self.assertTrue(Xe[0] - -3.5534271704166667 <= 1e-4)
        self.assertTrue(
            (
                np.array(Xf[0])
                - np.array(
                    [
                        [-0.03578407, -0.00176274, 0.03967705],
                        [0.06499199, -0.02277058, 0.03514361],
                        [-0.06499199, 0.02277058, -0.03514361],
                        [0.03578407, 0.00176274, -0.03967705],
                        [-0.03578407, 0.00176274, 0.03967705],
                        [0.06499199, 0.02277058, 0.03514361],
                        [-0.06499199, -0.02277058, -0.03514361],
                        [0.03578407, -0.00176274, -0.03967705],
                        [0.0, 0.02825717, 0.0],
                        [0.0, 0.02280701, 0.0],
                        [-0.02465965, -0.00049394, -0.04887495],
                        [-0.02165852, 0.02248651, -0.04979937],
                        [-0.01744417, -0.02081133, 0.03051514],
                        [0.01744417, 0.02081133, -0.03051514],
                        [0.02165852, -0.02248651, 0.04979937],
                        [0.02465965, 0.00049394, 0.04887495],
                        [0.0, -0.02280701, 0.0],
                        [0.0, -0.02825717, 0.0],
                        [-0.02465965, 0.00049394, -0.04887495],
                        [-0.02165852, -0.02248651, -0.04979937],
                        [-0.01744417, 0.02081133, 0.03051514],
                        [0.01744417, -0.02081133, -0.03051514],
                        [0.02165852, 0.02248651, 0.04979937],
                        [0.02465965, -0.00049394, 0.04887495],
                    ]
                )
            ).max()
            <= 1e-4
        )

    # def test_train(self):
    #    X_struct, Xe, Xf, Xs = self.agent.reverse_calcs(self.test_data['calcs_reversed'].iloc[:2])
    #    self.agent.train(self.seed_data, epochs=1)

    def test_hypotheses(self):
        candidate = self.agent.get_hypotheses(
            candidate_data=self.candidate_data, seed_data=self.seed_data, retrain=False
        )
        self.assertTrue(list(candidate.columns) == list(self.candidate_data.columns))


class StabilityAgentsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        test_data = pd.read_pickle(os.path.join(CAMD_TEST_FILES, "test_m3gnet.pickle"))
        print(test_data.columns)
        cls.test_data = test_data
        cls.seed_data, cls.candidate_data = train_test_split(
            test_data, train_size=0.2, random_state=42
        )
        model = M3GNet(is_intensive=False)
        cls.agent = M3GNetStabilityAgent(m3gnet=model)

    def test_m3gnet_agent(self):
        agent = M3GNetStabilityAgent()
        hypotheses = agent.get_hypotheses(
            candidate_data=self.candidate_data, seed_data=self.seed_data
        )


if __name__ == "__main__":
    unittest.main()
